import unittest
import mock
import logging
import tests.utils as utils
from tests.utils import compare_json_return_errormessage as error_message

# Import of code to be tested:
import esgfpid.assistant.publish
from esgfpid.exceptions import ArgumentError

# Logging
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import *
import resources.TESTVALUES as TESTHELPERS
from resources.TESTVALUES import TEST_RABBIT_CREDS_OPEN
from resources.TESTVALUES import TEST_RABBIT_CREDS_TRUSTED

# Some tests rely on open nodes
import tests.globalvar
import globalvar
if globalvar.RABBIT_OPEN_NOT_ALLOWED:
    print('Skipping tests that need open RabbitMQ nodes in module "%s".' % __name__)

class ConnectorTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    #
    # Init
    #

    '''
    Test the constructor, with trusted (and open) node.

    If open nodes are allowed, this should work fine.
    Otherwise, we expect an exception.
    '''
    @unittest.skipIf(not(globalvar.RABBIT_OPEN_NOT_ALLOWED), '(this test cannot cope with open rabbit nodes)')
    def test_init_trusted_and_open_ok_1(self):

        # Preparations: Connector args.
        # Use trusted and open node:
        rabbit_creds = [copy.deepcopy(TEST_RABBIT_CREDS_TRUSTED), copy.deepcopy(TEST_RABBIT_CREDS_OPEN)]
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = rabbit_creds
        )

        # Run code to be tested and check exception:
        with self.assertRaises(ArgumentError) as e:
            testconnector = esgfpid.Connector(**args)

        # Check result: Error message ok?
        self.assertIn('Missing password', str(e.exception))

    '''
    Test the constructor, with trusted (and open) node.

    If open nodes are allowed, this should work fine.
    Otherwise, we expect an exception.
    '''
    @unittest.skipIf(globalvar.RABBIT_OPEN_NOT_ALLOWED, '(this test uses open rabbit nodes)')
    def test_init_trusted_and_open_ok_2(self):

        # Preparations: Connector args.
        # Use trusted and open node:
        rabbit_creds = [copy.deepcopy(TEST_RABBIT_CREDS_TRUSTED), copy.deepcopy(TEST_RABBIT_CREDS_OPEN)]
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = rabbit_creds
        )

        # Run code to be tested: Connector constructior
        testconnector = esgfpid.Connector(**args)

        # Check result: Did init work?
        self.assertIsInstance(testconnector, esgfpid.Connector)

        # Check result: Did the module get the right number of
        # trusted and open rabbit nodes?
        node_manager = testconnector._Connector__coupler._Coupler__rabbit_message_sender._RabbitMessageSender__node_manager
        self.assertEqual(node_manager.get_num_left_trusted(), 1)
        self.assertEqual(node_manager.get_num_left_open(),1)


    def test_init_no_prefix(self):

        # Preparations: Connector args.
        rabbit_creds = [copy.deepcopy(TEST_RABBIT_CREDS_TRUSTED)]
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = rabbit_creds,
            handle_prefix = None
        )

        # Run code to be tested and check exception:
        with self.assertRaises(ArgumentError) as e:
            testconnector = esgfpid.Connector(**args)

        # Check result: Error message ok?
        self.assertIn('Missing handle prefix', str(e.exception))

    def test_init_wrong_prefix(self):

        # Preparations: Connector args.
        rabbit_creds = [copy.deepcopy(TEST_RABBIT_CREDS_TRUSTED)]
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = rabbit_creds,
            handle_prefix = '987654321'
        )

        # Run code to be tested and check exception:
        with self.assertRaises(ArgumentError) as e:
            testconnector = esgfpid.Connector(**args)

        # Check result: Error message ok?
        self.assertIn('is not a valid prefix', str(e.exception))

    def test_init_no_rabbit_url(self):

        # Preparations: Connector args.
        rabbit_creds = dict(
            user = RABBIT_USER_TRUSTED,
            password = RABBIT_PASSWORD
        )
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = [rabbit_creds]
        )

        # Run code to be tested and check exception:
        with self.assertRaises(ArgumentError) as e:
            testconnector = esgfpid.Connector(**args)

        # Check result: Error message ok?
        self.assertIn('Missing url for', str(e.exception))

    def test_init_no_rabbit_user(self):

        # Preparations: Connector args.
        rabbit_creds = dict(
            url = RABBIT_URL_TRUSTED,
            password = RABBIT_PASSWORD
        )
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = [rabbit_creds]
        )

        # Run code to be tested and check exception:
        with self.assertRaises(ArgumentError) as e:
            testconnector = esgfpid.Connector(**args)

        # Check result: Error message ok?
        self.assertIn('Missing user', str(e.exception))

    '''
    Test the constructor, with trusted node.
    '''
    def test_init_trusted_only_ok(self):

        # Preparation: Connector args.
        # Use trusted node:
        rabbit_creds = [copy.deepcopy(TEST_RABBIT_CREDS_TRUSTED)]
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = rabbit_creds
        )

        # Run code to be tested: Connector constructor
        testconnector = esgfpid.Connector(**args)

        # Check results: Did init work?
        self.assertIsInstance(testconnector, esgfpid.Connector)

        # Check results: Did the module get the right number of
        # trusted and open rabbit nodes?
        node_manager = testconnector._Connector__coupler._Coupler__rabbit_message_sender._RabbitMessageSender__node_manager
        self.assertEqual(node_manager.get_num_left_trusted(), 1)
        self.assertEqual(node_manager.get_num_left_open(),0)

    '''
    Test the constructor, with trusted node.
    '''
    def test_init_trusted_only_more_args_ok(self):

        # Preparation: Connector args.
        # Use trusted node:
        rabbit_creds = TESTHELPERS.get_rabbit_credentials(vhost='foo', port=666, ssl_enabled=True)
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = [rabbit_creds]
        )

        # Run code to be tested: Connector constructor
        testconnector = esgfpid.Connector(**args)

        # Check results: Did init work?
        self.assertIsInstance(testconnector, esgfpid.Connector)

        # Check results: Did the module get the right number of
        # trusted and open rabbit nodes?
        node_manager = testconnector._Connector__coupler._Coupler__rabbit_message_sender._RabbitMessageSender__node_manager
        self.assertEqual(node_manager.get_num_left_trusted(), 1)
        self.assertEqual(node_manager.get_num_left_open(),0)
        # Check: Were the right values passed?
        node_manager.set_next_host()
        curr = node_manager._NodeManager__current_node
        self.assertEqual(curr['vhost'],'foo')
        self.assertEqual(curr['port'],666)
        self.assertEqual(curr['ssl_enabled'],True)

    '''
    Test the constructor, with only open nodes.
    '''
    @unittest.skipIf(not(globalvar.RABBIT_OPEN_NOT_ALLOWED), '(this test cannot deal with open rabbit nodes)')
    def test_init_open_ok(self):

        # Preparation: Connector args.
        # Use open node:
        rabbit_creds = [copy.deepcopy(TEST_RABBIT_CREDS_OPEN)]
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = rabbit_creds
        )

        # Run code to be tested and check exception:
        with self.assertRaises(ArgumentError) as e:
            testconnector = esgfpid.Connector(**args)

        # Check result: Error message ok?
        self.assertIn('Missing password', str(e.exception))


    '''
    Test the constructor, with only open nodes.
    '''
    @unittest.skipIf(globalvar.RABBIT_OPEN_NOT_ALLOWED, '(this test uses open rabbit nodes)')
    def test_init_open_ok_2(self):

        # Preparation: Connector args.
        # Use open node:
        rabbit_creds = [copy.deepcopy(TEST_RABBIT_CREDS_OPEN)]
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = rabbit_creds
        )

        # Run code to be tested: Connector constructor
        testconnector = esgfpid.Connector(**args)

        # Check results: Did init work?
        self.assertIsInstance(testconnector, esgfpid.Connector)

        # Check results: Did the module get the right number of
        # trusted and open rabbit nodes?
        node_manager = testconnector._Connector__coupler._Coupler__rabbit_message_sender._RabbitMessageSender__node_manager
        self.assertEqual(node_manager.get_num_left_trusted(), 0)
        self.assertEqual(node_manager.get_num_left_open(),1)

    def test_init_rabbit_user_as_list(self):

        # Preparations: Connector args.
        rabbit_creds = dict(
            url = RABBIT_URL_TRUSTED,
            user = [RABBIT_USER_TRUSTED],
            password = RABBIT_PASSWORD
        )
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = [rabbit_creds]
        )

        # Run code to be tested:
        testconnector = esgfpid.Connector(**args)

        # Check result: Did init work?
        self.assertIsInstance(testconnector, esgfpid.Connector)

    def test_init_too_many_rabbit_users(self):

        # Preparations: Connector args.
        rabbit_creds = dict(
            url = RABBIT_URL_TRUSTED,
            user = [RABBIT_USER_TRUSTED, 'johndoe', 'alicedoe'],
            password = RABBIT_PASSWORD
        )
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = [rabbit_creds]
        )

        # Run code to be tested:
        with self.assertRaises(ArgumentError) as e:
            testconnector = esgfpid.Connector(**args)

        # Check result: Error message ok?
        self.assertIn('Wrong type', str(e.exception))

    def test_init_vhost_no_string(self):

        # Preparations: Connector args.
        rabbit_creds = dict(
            url = RABBIT_URL_TRUSTED,
            user = RABBIT_USER_TRUSTED,
            password = RABBIT_PASSWORD,
            vhost = 123
        )
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = [rabbit_creds]
        )

        # Run code to be tested:
        with self.assertRaises(ArgumentError) as e:
            testconnector = esgfpid.Connector(**args)

        # Check result: Error message ok?
        self.assertIn('Wrong type', str(e.exception))

    def test_init_sslenabled_no_bool(self):

        # Preparations: Connector args.
        rabbit_creds = dict(
            url = RABBIT_URL_TRUSTED,
            user = RABBIT_USER_TRUSTED,
            password = RABBIT_PASSWORD,
            ssl_enabled = 123
        )
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = [rabbit_creds]
        )

        # Run code to be tested:
        with self.assertRaises(ArgumentError) as e:
            testconnector = esgfpid.Connector(**args)

        # Check result: Error message ok?
        self.assertIn('Wrong type', str(e.exception))

    def test_init_sslenabled_string_bool_true(self):

        # Preparations: Connector args.
        rabbit_creds = dict(
            url = RABBIT_URL_TRUSTED,
            user = RABBIT_USER_TRUSTED,
            password = RABBIT_PASSWORD,
            ssl_enabled = 'tRuE'
        )
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = [rabbit_creds]
        )

        # Run code to be tested:
        testconnector = esgfpid.Connector(**args)

        # Check result: Did init work?
        self.assertIsInstance(testconnector, esgfpid.Connector)
        # Check: Were the right values passed?
        node_manager = testconnector._Connector__coupler._Coupler__rabbit_message_sender._RabbitMessageSender__node_manager
        node_manager.set_next_host()
        curr = node_manager._NodeManager__current_node
        self.assertEqual(curr['ssl_enabled'],True)

    def test_init_sslenabled_string_bool_false(self):

        # Preparations: Connector args.
        rabbit_creds = dict(
            url = RABBIT_URL_TRUSTED,
            user = RABBIT_USER_TRUSTED,
            password = RABBIT_PASSWORD,
            ssl_enabled = 'fAlSe'
        )
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = [rabbit_creds]
        )

        # Run code to be tested:
        testconnector = esgfpid.Connector(**args)

        # Check result: Did init work?
        self.assertIsInstance(testconnector, esgfpid.Connector)
        # Check: Were the right values passed?
        node_manager = testconnector._Connector__coupler._Coupler__rabbit_message_sender._RabbitMessageSender__node_manager
        node_manager.set_next_host()
        curr = node_manager._NodeManager__current_node
        self.assertEqual(curr['ssl_enabled'],False)

    def test_init_sslenabled_string_bool_other(self):

        # Preparations: Connector args.
        rabbit_creds = dict(
            url = RABBIT_URL_TRUSTED,
            user = RABBIT_USER_TRUSTED,
            password = RABBIT_PASSWORD,
            ssl_enabled = 'yes'
        )
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = [rabbit_creds]
        )

        # Run code to be tested:
        with self.assertRaises(ArgumentError) as e:
            testconnector = esgfpid.Connector(**args)

        # Check result: Error message ok?
        self.assertIn('Wrong type', str(e.exception))

    def test_init_sslenabled_string_bool_other(self):

        # Preparations: Connector args.
        rabbit_creds = dict(
            url = RABBIT_URL_TRUSTED,
            user = RABBIT_USER_TRUSTED,
            password = RABBIT_PASSWORD,
            ssl_enabled = '',
            vhost = '',
            port = ''
        )
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = [rabbit_creds]
        )

        # Run code to be tested:
        testconnector = esgfpid.Connector(**args)

        # Check result: Did init work?
        self.assertIsInstance(testconnector, esgfpid.Connector)
        # Check: Were the right values passed?
        node_manager = testconnector._Connector__coupler._Coupler__rabbit_message_sender._RabbitMessageSender__node_manager
        node_manager.set_next_host()
        curr = node_manager._NodeManager__current_node
        self.assertEqual(curr['ssl_enabled'],None)
        self.assertEqual(curr['vhost'],None)
        self.assertEqual(curr['port'],None)

    def test_init_port_no_int(self):

        # Preparations: Connector args.
        rabbit_creds = dict(
            url = RABBIT_URL_TRUSTED,
            user = RABBIT_USER_TRUSTED,
            password = RABBIT_PASSWORD,
            port = 'foo'
        )
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = [rabbit_creds]
        )

        # Run code to be tested:
        with self.assertRaises(ArgumentError) as e:
            testconnector = esgfpid.Connector(**args)

        # Check result: Error message ok?
        self.assertIn('Wrong type', str(e.exception))

    def test_init_port_string_int(self):

        # Preparations: Connector args.
        rabbit_creds = dict(
            url = RABBIT_URL_TRUSTED,
            user = RABBIT_USER_TRUSTED,
            password = RABBIT_PASSWORD,
            port = '123'
        )
        args = TESTHELPERS.get_connector_args(
            messaging_service_credentials = [rabbit_creds]
        )

        # Run code to be tested:
        testconnector = esgfpid.Connector(**args)

        # Check result: Did init work?
        self.assertIsInstance(testconnector, esgfpid.Connector)
        # Check: Were the right values passed?
        node_manager = testconnector._Connector__coupler._Coupler__rabbit_message_sender._RabbitMessageSender__node_manager
        node_manager.set_next_host()
        curr = node_manager._NodeManager__current_node
        self.assertEqual(curr['ssl_enabled'],None)
        self.assertEqual(curr['vhost'],None)
        self.assertEqual(curr['port'],123)

    '''
    Test if the solr URL is passed to the consumer in the
    message.
    (The client wants to tell the consumer which solr
    instance to use for looking up dataset versions.
    Consumer has a default, if it is not passed. Consumer
    does not necessarily use solr, it might also use the
    handle database.)
    '''
    def test_passing_consumer_solr_url_ok(self):

        # Run code to be tested: Create connector with consumer solr url:
        # (And with the two necessary args for dataset publication)
        args = TESTHELPERS.get_connector_args(
            consumer_solr_url='fake_solr_whatever',
            thredds_service_path='foo',
            data_node='bar'
        )
        testconnector = esgfpid.Connector(**args)

        # Run code to be tested: Init dataset wizard
        wizard = testconnector.create_publication_assistant(
            drs_id='baz',
            version_number=2016,
            is_replica=False
        )

        # Check result: Did init work?
        self.assertIsInstance(testconnector, esgfpid.Connector)
        self.assertIsInstance(wizard, esgfpid.assistant.publish.DatasetPublicationAssistant)

        # Check result: Was the consumer solr url passed?
        self.assertIsNotNone(testconnector._Connector__consumer_solr_url)
        self.assertIsNotNone(wizard._DatasetPublicationAssistant__consumer_solr_url)


    '''
    Test whether the correct defaults are set
    if the connector is initialized with the minimum
    arguments.

    '''
    def test_init_default_args_to_coupler(self):

        # Preparations: Minimum args for connector
        args = TESTHELPERS.get_connector_args()

        # Run code to be tested: Connector constructor
        testconnector = esgfpid.Connector(**args)

        # Check results: Check if the correct defaults
        # were set (i.e. passed to coupler.)
        coupler_args = testconnector._Connector__coupler.args
        self.assertEqual(coupler_args['data_node'],None)
        self.assertEqual(coupler_args['thredds_service_path'],None)
        self.assertEqual(coupler_args['test_publication'],False)
        self.assertEqual(coupler_args['solr_url'],None)
        self.assertEqual(coupler_args['solr_switched_off'],True)
        self.assertEqual(coupler_args['solr_https_verify'],False)
        self.assertEqual(coupler_args['disable_insecure_request_warning'],False)
        self.assertEqual(coupler_args['message_service_synchronous'],False)
        self.assertEqual(coupler_args['consumer_solr_url'],None)
        
    '''
    Test whether the correct defaults are set
    if the connector is initialized with the minimum
    arguments.
    '''
    def test_init_not_default_args_to_coupler(self):

        # Run code to be tested: Make a connector
        # with many non-default args:
        args = TESTHELPERS.get_connector_args(
            data_node=DATA_NODE,
            thredds_service_path=THREDDS,
            test_publication=True,
            solr_url=SOLR_URL_LIBRARY,
            solr_https_verify=True,
            disable_insecure_request_warning=True,
            message_service_synchronous=True,
            consumer_solr_url=SOLR_URL_CONSUMER
        )
        testconnector = esgfpid.Connector(**args)

        # Check results: Check if the correct defaults
        # were set (i.e. passed to coupler.)
        coupler_args = testconnector._Connector__coupler.args
        self.assertEqual(coupler_args['data_node'],DATA_NODE)
        self.assertEqual(coupler_args['thredds_service_path'],THREDDS)
        self.assertEqual(coupler_args['test_publication'],True)
        self.assertEqual(coupler_args['solr_url'],SOLR_URL_LIBRARY)
        self.assertEqual(coupler_args['solr_switched_off'],False)
        self.assertEqual(coupler_args['solr_https_verify'],True)
        self.assertEqual(coupler_args['disable_insecure_request_warning'],True)
        self.assertEqual(coupler_args['message_service_synchronous'],True)
        self.assertEqual(coupler_args['consumer_solr_url'],SOLR_URL_CONSUMER)

    '''
    Check if solr is not switched off if an URL given.
    '''
    def test_init_solr_not_off(self):

        # Run code to be tested: Make connector with solr url
        args = TESTHELPERS.get_connector_args(solr_url='foo')
        testconnector = esgfpid.Connector(**args)

        # Check results: Check if the correct defaults
        # were set (i.e. passed to coupler.)
        coupler_args = testconnector._Connector__coupler.args
        self.assertEqual(coupler_args['solr_url'],'foo')
        self.assertEqual(coupler_args['solr_switched_off'],False)

    '''
    Check if solr is not switched off if an URL given.
    '''
    def test_init_solr_off(self):

        # Preparations: Minimum args for connector
        args = TESTHELPERS.get_connector_args()

        # Run code to be tested: Connector constructor
        testconnector = esgfpid.Connector(**args)

        # Check results: Check if the correct defaults
        # were set (i.e. passed to coupler.)
        coupler_args = testconnector._Connector__coupler.args
        self.assertEqual(coupler_args['solr_url'],None)
        self.assertEqual(coupler_args['solr_switched_off'],True)


    #
    # Publication
    #

    '''
    This passes the correct arguments.
    '''
    def test_create_publication_assistant_ok(self):

        # Preparations: Create connector with data node and thredds:
        args = TESTHELPERS.get_connector_args(
            thredds_service_path='foo',
            data_node='bar'
        )
        testconnector = esgfpid.Connector(**args)

        # Run code to be tested: Init dataset wizard
        wizard = testconnector.create_publication_assistant(
            drs_id='baz',
            version_number=2016,
            is_replica=False
        )

        # Check result:
        self.assertIsInstance(wizard, esgfpid.assistant.publish.DatasetPublicationAssistant)


    '''
    If we want to publish a dataset, "data_node" and
    "thredds_service_path" have to be specified in the beginning!
    '''
    def test_create_publication_assistant_missing_thredds(self):

        # Preparations: Make connector without thredds:
        args = TESTHELPERS.get_connector_args(data_node = 'foo')
        testconnector = esgfpid.Connector(**args)

        # Run code to be tested and check exception: Init dataset wizard
        with self.assertRaises(esgfpid.exceptions.ArgumentError) as e:
            wizard = testconnector.create_publication_assistant(
                drs_id='bar',
                version_number=2016,
                is_replica=False
            )

        # Check result: Error message ok?
        self.assertIn('No thredds_service_path given', str(e.exception))

    '''
    If we want to publish a dataset, "data_node" and
    "thredds_service_path" have to be specified in the beginning!
    '''
    def test_create_publication_assistant_missing_datanode(self):

        # Preparations: Make connector without thredds:
        args = TESTHELPERS.get_connector_args(thredds_service_path = 'foo')
        testconnector = esgfpid.Connector(**args)

        # Run code to be tested and check exception: Init dataset wizard
        with self.assertRaises(esgfpid.exceptions.ArgumentError) as e:
            wizard = testconnector.create_publication_assistant(
                drs_id='bar',
                version_number=2016,
                is_replica=False
            )

        # Check result: Error message ok?
        self.assertIn('No data_node given', str(e.exception))

    #
    # Unpublication
    #

    '''
    If we want to unpublish a dataset, "data_node" has to
    be specified in the beginning!
    '''
    def test_unpublish_all_versions_missing_data_node(self):

        # Preparations: Make patched connector without the
        # necessary data node (needed for unpublish)
        testconnector = TESTHELPERS.get_connector()
        TESTHELPERS.patch_with_rabbit_mock(testconnector)

        # Run code to be tested: Unpublish
        with self.assertRaises(esgfpid.exceptions.ArgumentError) as e:
            testconnector.unpublish_all_versions(drs_id=DRS_ID)

        # Check result: Error message ok?
        self.assertIn('No data_node given', str(e.exception))

    '''
    This passes the correct args.
    '''
    def test_unpublish_one_version_ok(self):

        # Preparations: Make patched connector with data node (needed for unpublish)
        testconnector = TESTHELPERS.get_connector(data_node=DATA_NODE)
        TESTHELPERS.patch_with_rabbit_mock(testconnector)
  
        # Run code to be tested: Unpublish
        testconnector.unpublish_one_version(
            drs_id=DRS_ID,
            version_number=DS_VERSION
        )

        # Check result:
        expected_rabbit_task = {
            "handle": DATASETHANDLE_HDL,
            "operation": "unpublish_one_version",
            "message_timestamp":"anydate",
            "aggregation_level":"dataset",
            "data_node": DATA_NODE,
            "ROUTING_KEY":PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.unpubli-onevers',
            "drs_id":DRS_ID,
            "version_number": int(DS_VERSION)
        }
        received_rabbit_msg = TESTHELPERS.get_received_message_from_rabbitmock(testconnector)
        is_same = utils.is_json_same(expected_rabbit_task, received_rabbit_msg)
        self.assertTrue(is_same, utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_msg))

    '''
    This passes the correct args.
    '''
    def test_unpublish_one_version_by_handle_ok(self):

        # Preparations: Make patched connector with data node (needed for unpublish)
        testconnector = TESTHELPERS.get_connector(data_node=DATA_NODE)
        TESTHELPERS.patch_with_rabbit_mock(testconnector)
  
        # Run code to be tested: Unpublish
        testconnector.unpublish_one_version(
            dataset_handle = DATASETHANDLE_HDL
        )

        # Check result:
        expected_rabbit_task = {
            "handle": DATASETHANDLE_HDL,
            "operation": "unpublish_one_version",
            "message_timestamp":"anydate",
            "aggregation_level":"dataset",
            "data_node": DATA_NODE,
            "ROUTING_KEY":PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.unpubli-onevers',
            "drs_id": None,
        }
        received_rabbit_msg = TESTHELPERS.get_received_message_from_rabbitmock(testconnector)
        is_same = utils.is_json_same(expected_rabbit_task, received_rabbit_msg)
        self.assertTrue(is_same, utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_msg))

    '''
    We unpublish one version.

    For this, no solr url is needed, as the version is already
    specified. So the consumer_solr_url is not passed on.
    '''
    def test_unpublish_one_version_with_consumer_url_ok(self):

        # Preparations: Make connector, but without
        # solr mock. This is to see that with consumer_solr_url,
        # the library does not try to access solr.
        testconnector = TESTHELPERS.get_connector(
            consumer_solr_url=SOLR_URL_CONSUMER,
            data_node=DATA_NODE
        )
        TESTHELPERS.patch_with_rabbit_mock(testconnector)
  
        # Run code to be tested: Unpublish one version
        testconnector.unpublish_one_version(
            drs_id=DRS_ID,
            version_number=DS_VERSION
        )            

        # Check result:
        # We don't get the consumer_solr_url, because it is only
        # needed for unpublishing all versions.
        expected_rabbit_task = {
            "handle": DATASETHANDLE_HDL,
            "operation": "unpublish_one_version",
            "message_timestamp":"anydate",
            "aggregation_level":"dataset",
            "data_node": DATA_NODE,
            "ROUTING_KEY":PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.unpubli-onevers',
            "drs_id":DRS_ID,
            "version_number": int(DS_VERSION)
        }
        received_rabbit_msg = TESTHELPERS.get_received_message_from_rabbitmock(testconnector)
        is_same = utils.is_json_same(expected_rabbit_task, received_rabbit_msg)
        self.assertTrue(is_same, utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_msg))

    '''
    Test unpublishing all versions.

    For this, the library would call solr to find handles of
    all versions. We make our mock solr return None, so the
    library must tell the consumer to look for the versions
    itself.

    This time, we tell the consumer to use its default solr url.
    '''
    @mock.patch('esgfpid.coupling.Coupler.retrieve_datasethandles_or_versionnumbers_of_allversions')
    def test_unpublish_all_versions_nosolr_ok(self, solr_asker_patch):

        # Patch: Make the coupler return None when we ask it
        # to find dataset versions...
        mydict = dict(dataset_handles=None, version_numbers=None)
        solr_asker_patch.return_value = mydict

        # Preparations: Create patched connector
        testconnector = TESTHELPERS.get_connector(data_node=DATA_NODE)
        TESTHELPERS.patch_with_rabbit_mock(testconnector)

        # Run code to be tested:
        testconnector.unpublish_all_versions(drs_id=DRS_ID)

        # Check result:
        expected_rabbit_task = {
            "operation": "unpublish_all_versions",
            "message_timestamp": "anydate",
            "data_node": DATA_NODE,
            "aggregation_level":"dataset",
            "drs_id":DRS_ID,
            "ROUTING_KEY":PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.unpubli-allvers',
        }
        received_rabbit_msg = TESTHELPERS.get_received_message_from_rabbitmock(testconnector)
        is_same = utils.is_json_same(expected_rabbit_task, received_rabbit_msg)
        self.assertTrue(is_same, utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_msg))

    def test_unpublish_all_versions_solr_off(self):

        # Preparations: Create patched connector
        testconnector = TESTHELPERS.get_connector(
            data_node=DATA_NODE,
            solr_switched_off=True 
        )
        TESTHELPERS.patch_with_rabbit_mock(testconnector)

        # Run code to be tested:
        testconnector.unpublish_all_versions(drs_id=DRS_ID)

        # Check result:
        expected_rabbit_task = {
            "operation": "unpublish_all_versions",
            "message_timestamp": "anydate",
            "data_node": DATA_NODE,
            "aggregation_level":"dataset",
            "drs_id":DRS_ID,
            "ROUTING_KEY":PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.unpubli-allvers',
        }
        received_rabbit_msg = TESTHELPERS.get_received_message_from_rabbitmock(testconnector)
        is_same = utils.is_json_same(expected_rabbit_task, received_rabbit_msg)
        self.assertTrue(is_same, utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_msg))


    '''
    Test unpublishing all versions.


    For this, the library would call solr to find handles of
    all versions. We make our mock solr return None, so the
    library must tell the consumer to look for the versions
    itself.

    This time, we tell the consumer to use the solr url
    we provide.
    '''
    @mock.patch('esgfpid.coupling.Coupler.retrieve_datasethandles_or_versionnumbers_of_allversions')
    def test_unpublish_all_versions_nosolr__butconsumersolr_ok(self, solr_asker_patch):

        # Patch coupler
        mydict = dict(dataset_handles=None, version_numbers=None)
        solr_asker_patch.return_value = mydict

        # Preparations: Create patched connector
        testconnector = TESTHELPERS.get_connector(
            data_node = DATA_NODE,
            solr_url = SOLR_URL_LIBRARY,
            consumer_solr_url=SOLR_URL_CONSUMER
        )
        TESTHELPERS.patch_with_rabbit_mock(testconnector)

        # Run code to be tested:
        testconnector.unpublish_all_versions(
            drs_id=DRS_ID,
            data_node=DATA_NODE)

        # Check result:
        expected_rabbit_task = {
            "operation": "unpublish_all_versions",
            "message_timestamp": "anydate",
            "data_node": DATA_NODE,
            "aggregation_level":"dataset",
            "drs_id":DRS_ID,
            "ROUTING_KEY":PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.unpubli-allvers',
            "consumer_solr_url":SOLR_URL_CONSUMER
        }        
        received_rabbit_msg = TESTHELPERS.get_received_message_from_rabbitmock(testconnector)
        is_same = utils.is_json_same(expected_rabbit_task, received_rabbit_msg)
        self.assertTrue(is_same, utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_msg))

    #
    # Errata
    #

    def test_add_errata_id_several_ok(self):

        # Preparations: Create patched connector
        # (Patched to avoid that message be sent, and to retrieve the created message)
        testconnector = TESTHELPERS.get_connector()
        TESTHELPERS.patch_with_rabbit_mock(testconnector)

        # Run code to be tested: Add errata
        testconnector.add_errata_ids(
            drs_id=DRS_ID,
            version_number=DS_VERSION,
            errata_ids=ERRATA_SEVERAL
        )

        # Check result: Was correct errata message created?
        expected_rabbit_task = {
            "handle": DATASETHANDLE_HDL,
            "operation": "add_errata_ids",
            "message_timestamp":"anydate",
            "errata_ids":ERRATA_SEVERAL,
            "ROUTING_KEY":PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.errata-add',
            "drs_id":DRS_ID,
            "version_number":DS_VERSION
        }
        received_rabbit_msg = TESTHELPERS.get_received_message_from_rabbitmock(testconnector)
        is_same = utils.is_json_same(expected_rabbit_task, received_rabbit_msg)
        self.assertTrue(is_same, utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_msg))

    def test_add_errata_id_one_ok(self):

        # Preparations: Create patched connector
        # (Patched to avoid that message be sent, and to retrieve the created message)
        testconnector = TESTHELPERS.get_connector()
        TESTHELPERS.patch_with_rabbit_mock(testconnector)
        
        # Run code to be tested: Add errata
        testconnector.add_errata_ids(
            drs_id=DRS_ID,
            version_number=DS_VERSION,
            errata_ids=ERRATA
        )

        # Check result: Was correct errata message created?
        expected_rabbit_task = {
            "handle": DATASETHANDLE_HDL,
            "operation": "add_errata_ids",
            "message_timestamp":"anydate",
            "errata_ids":[ERRATA],
            "ROUTING_KEY":PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.errata-add',
            "drs_id":DRS_ID,
            "version_number":DS_VERSION
        }
        received_rabbit_msg = TESTHELPERS.get_received_message_from_rabbitmock(testconnector)
        is_same = utils.is_json_same(expected_rabbit_task, received_rabbit_msg)
        self.assertTrue(is_same, utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_msg))

    def test_remove_errata_id_one_ok(self):

        # Preparations: Create patched connector
        # (Patched to avoid that message be sent, and to retrieve the created message)
        testconnector = TESTHELPERS.get_connector()
        TESTHELPERS.patch_with_rabbit_mock(testconnector)

        # Run code to be tested: Add errata
        testconnector.remove_errata_ids(
            drs_id=DRS_ID,
            version_number=DS_VERSION,
            errata_ids=ERRATA
        )

        # Check result: Was correct errata message created?
        expected_rabbit_task = {
            "handle": DATASETHANDLE_HDL,
            "operation": "remove_errata_ids",
            "message_timestamp":"anydate",
            "errata_ids":[ERRATA],
            "ROUTING_KEY":PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.errata-rem',
            "drs_id":DRS_ID,
            "version_number":DS_VERSION
        }
        received_rabbit_msg = TESTHELPERS.get_received_message_from_rabbitmock(testconnector)
        is_same = utils.is_json_same(expected_rabbit_task, received_rabbit_msg)
        self.assertTrue(is_same, utils.compare_json_return_errormessage(expected_rabbit_task, received_rabbit_msg))


    #
    # Data Cart
    #

    def test_make_data_cart_pid(self):

        # Test variables
        content1 = {'foo':'foo', 'bar':'bar'}
        content2 = {'foo':'foo', 'bar': None}

        # Preparations: Create patched connector
        # (Patched to avoid that message be sent, and to retrieve the created message)
        testconnector = TESTHELPERS.get_connector()
        TESTHELPERS.patch_with_rabbit_mock(testconnector)

        # Run code to be tested: Create data cart PIDs
        # And retrieve the messages
        pid1 = testconnector.create_data_cart_pid(content1)
        pid2 = testconnector.create_data_cart_pid(content2)

        # Check result: Were the correct messages created?
        expected_handle_both_cases = PREFIX_WITH_HDL+"/b597a79e-1dc7-3d3f-b689-75ac5a78167f"
        expected_rabbit_task1 = {
            "handle": expected_handle_both_cases,
            "operation": "shopping_cart",
            "message_timestamp":"anydate",
            "data_cart_content":content1,
            "ROUTING_KEY":PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.datacart'
        }
        expected_rabbit_task2 = {
            "handle": expected_handle_both_cases,
            "operation": "shopping_cart",
            "message_timestamp":"anydate",
            "data_cart_content":content2,
            "ROUTING_KEY":PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.datacart'
        }
        received_rabbit_msg1 = TESTHELPERS.get_received_message_from_rabbitmock(testconnector, 0)
        received_rabbit_msg2 = TESTHELPERS.get_received_message_from_rabbitmock(testconnector, 1)
        same1 = utils.is_json_same(expected_rabbit_task1, received_rabbit_msg1)
        same2 = utils.is_json_same(expected_rabbit_task2, received_rabbit_msg2)
        self.assertTrue(same1, error_message(expected_rabbit_task1, received_rabbit_msg1))
        self.assertTrue(same2, error_message(expected_rabbit_task2, received_rabbit_msg2))
        self.assertTrue(pid1==pid2, 'Both pids are not the same.')

    #
    # Threads
    #

    def test_start_messaging_thread(self):
        LOGGER.debug('Thread test')
        testconnector = TESTHELPERS.get_connector()
        rabbitmock = TESTHELPERS.patch_with_rabbit_mock(testconnector, mock.MagicMock())
        testconnector.start_messaging_thread()
        rabbitmock.start.assert_called_with()

    def test_finish_messaging_thread(self):
        LOGGER.debug('Thread test')
        testconnector = TESTHELPERS.get_connector()
        rabbitmock = TESTHELPERS.patch_with_rabbit_mock(testconnector, mock.MagicMock())
        testconnector.finish_messaging_thread()
        rabbitmock.finish.assert_called_with()

    def test_force_finish_messaging_thread(self):
        LOGGER.debug('Thread test')
        testconnector = TESTHELPERS.get_connector()
        rabbitmock = TESTHELPERS.patch_with_rabbit_mock(testconnector, mock.MagicMock())
        testconnector.force_finish_messaging_thread()
        rabbitmock.force_finish.assert_called_with()

    def test_make_handle(self):
        testconnector = TESTHELPERS.get_connector()
        received_handle = testconnector.make_handle_from_drsid_and_versionnumber(
            drs_id=DRS_ID,
            version_number=DS_VERSION
        )
        expected_handle = DATASETHANDLE_HDL
        self.assertEqual(received_handle, expected_handle)



import json
import random
import logging
import esgfpid.defaults
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

'''
Retrieves the routing key from the message, checks
the message's validity and returns
the string message and the routing key string.

If message is string, convert to JSON (dictionary).
If message is dictionary, check if it's valid JSON.
Retrieve the routing key, which should be included
with the key "ROUTING_KEY". Otherwise, a default
routing key is added.

:param msg: Message to be sent to RabbitMQ as JSON
    string or dictionary.
:return: Routing key string and string message as tuple.
'''
def get_routing_key_and_string_message_from_message_if_possible(msg):

    # Try to convert message to json:
    json_ok = False
    msg_json = None
    msg_string = None
    
    if msg is None:
        raise ValueError('The message that was passed is None.')
    
    # Get JSON from message, if possible!
    if isinstance(msg, str):

        try:
            # Valid string message --> JSON
            msg_string = msg
            msg_json = json.loads(msg)
            json_ok = True
            logdebug(LOGGER, 'Message was transformed to json.')
        except ValueError as e:
   
            # Try again, in case there was ' instead of "
            if "'" in msg:
                msg_string = msg
                msg_temp = msg.replace("'", '"')
                msg_json = json.loads(msg_temp)
                json_ok = True
                logdebug(LOGGER, 'Message was transformed to json (after replacing single quotes with double quotes).')

            else:
                # Invalid string message
                loginfo(LOGGER, 'Message seems to be invalid json: %s', msg)
                msg_string = str(msg)
                json_ok = False
    else:
        try:
            # Message is json already.
            msg_string = json.dumps(msg)
            msg_json = msg
            json_ok = True
            logtrace(LOGGER, 'Message was already json.')

        except TypeError as e:
            if 'not JSON serializable' in repr(e):

                # Message was whatever.
                msg_string = str(msg)
                json_ok = False
                msg = ('Message was neither JSON nor string and not understandable: %s' % msg_string)
                loginfo(LOGGER, msg)
                raise ValueError(msg)


    # If we succeeded, try to get routing key:
    routing_key = None
    if json_ok:
        try:
            routing_key = msg_json['ROUTING_KEY']
            logtrace(LOGGER, 'Routing key extracted from message.')
        except (KeyError, TypeError) as e:
            logdebug(LOGGER, 'No routing key in message.')
            routing_key = esgfpid.utils.RABBIT_DEFAULT_ROUTING_KEY
            pass # There is no routing key in the message
    else:
        routing_key = esgfpid.utils.RABBIT_DEFAULT_ROUTING_KEY

    return routing_key, msg_string


def add_emergency_routing_key(body_json):
    emergency_routing_key = esgfpid.utils.RABBIT_EMERGENCY_ROUTING_KEY

    # If it already HAS the emergency routing key, do not adapt the routing key
    # (This means the message already came back a second time...)
    if body_json[esgfpid.assistant.messages.JSON_KEY_ROUTING_KEY] == emergency_routing_key:
        pass

    # Otherwise, store the original one in another field...
    # and overwrite it by the emergency routing key:
    else:
        try:
            body_json['original_routing_key'] = body_json[esgfpid.assistant.messages.JSON_KEY_ROUTING_KEY]

        # If there was no routing key, set the original one to 'None'
        except KeyError:
            logerror('Very unexpected: RabbitMQ returned a message that had no routing key: %s', body_json)
            body_json['original_routing_key'] = 'None'

        logdebug(LOGGER, 'Adding emergency routing key %s', emergency_routing_key)
        body_json[esgfpid.assistant.messages.JSON_KEY_ROUTING_KEY] = emergency_routing_key
    return body_json, emergency_routing_key


# Default values for the esgfpid library

# List of prefixes
ACCEPTED_PREFIXES = ['21.14100','21.14101','21.14102','21.14103', '21.14104', '21.14foo']
# 21.14100: CMIP6, input4MIPs
# 21.14101: datacarts
# 21.14102: obs4MIPs Handles
# 21.14203: Cordex
# 21.14104: CMIP-Adjust

# Custom logging
LOG_INFO_TO_DEBUG=False # info messages are printed as "debug" to not appear on publisher console. Should be False, unless you want to almost silence it.
LOG_DEBUG_TO_INFO=False # debug messages are printed as "info" to appear on publisher console. Should be False, except for during tests.
LOG_TRACE_TO_DEBUG=False # Print extremely detailed log messages of rabbit module as "DEBUG" messages. Should be False, except for during tests.
LOG_SHOW_TO_INFO=False # So I can selectively show some log messages to Katharina without having to modify the whole code. Deprecated.
RABBIT_LOG_MESSAGE_INCREMENT = 10

# Solr:
SOLR_HTTPS_VERIFY_DEFAULT=False
SOLR_QUERY_DISTRIB=False

# Rabbit
RABBIT_IS_ASYNCHRONOUS = True
_persistent = 2
_nonpersistent = 1
RABBIT_DELIVERY_MODE = _persistent # 'delivery_mode': See https://pika.readthedocs.org/en/0.9.6/examples/comparing_publishing_sync_async.html#comparing-message-publishing-with-blockingconnection-and-selectconnection
RABBIT_MANDATORY_DELIVERY = True  # 'mandatory':  "This flag tells the server how to react if the message cannot be routed to a queue. If this flag is set, the server will return an unroutable message with a Return method. If this flag is zero, the server silently drops the message." # See: http://www.rabbitmq.com/amqp-0-9-1-reference.html#basic.publish
RABBIT_FALLBACK_EXCHANGE_NAME = "FALLBACK"

# Rabbit module, values for pika
# See http://pika.readthedocs.io/en/latest/_modules/pika/connection.html
RABBIT_PIKA_SOCKET_TIMEOUT=0.25 # defaults to 0.25 sec
RABBIT_PIKA_CONNECTION_ATTEMPTS=1 # defaults to 1
RABBIT_PIKA_CONNECTION_RETRY_DELAY_SECONDS=0 # defaults to 2.0
# Rabbit reconnection attempts
RABBIT_RECONNECTION_SECONDS=0.5 # after how much time try to retry connecting to same hosts if connection was closed?
RABBIT_RECONNECTION_MAX_TRIES=2 # how many times should the module try reconnecting? Not so many times, rather throw exception to publisher.
# Rabbit attempts to send message (synchronous only):
RABBIT_SYN_MESSAGE_MAX_TRIES=3
RABBIT_SYN_MESSAGE_TIMEOUT_MILLISEC=10
# Rabbit closing down algorithm (asynchronous only):
RABBIT_ASYN_FINISH_MAX_TRIES=10 # How many times to recheck if all messages are published+confirmed (on finish)
RABBIT_ASYN_FINISH_WAIT_SECONDS=0.5 # How much time to wait until you recheck (on finish)




import logging
import pika
import sys
if sys.version[0] == '2':
    import Queue as queue
else:
    import queue as queue
from .. import rabbitutils
import esgfpid.defaults as defaults
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn, log_every_x_times

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

'''
The RabbitFeeder is responsible for publishing messages to RabbitMQ.

It is very simple. Basically the only method it exposes
(except for some simple getter/setter which is rarely ever used)
is publish_message(), which is called from the main thread.

'''
class RabbitFeeder(object):

    def __init__(self, thread, statemachine, nodemanager):
        self.thread = thread

        '''
        Read-only.
        Before publishing a message, we check the state, and we log 
        the state. '''
        self.statemachine = statemachine

        self.nodemanager = nodemanager

        '''
        The deliver_number is important. It defines the number of the message
        that is used to identify it between this client and the RabbitMQ
        server (e.g. so the correct messages are deleted upon confirmation).

        It makes sure that rabbit server and this client talk about the
        same message.
        NEVER EVER INCREMENT OR OTHERWISE MODIFY THIS!

        From the RabbitMQ docs:
        "The delivery tag is valid only within the channel from which
        the message was received. I.e. a client MUST NOT receive a
        message on one channel and then acknowledge it on another."
        Source: https://www.rabbitmq.com/amqp-0-9-1-reference.html '''
        self.__delivery_number = 1

        # Logging
        self.__first_publication_trigger = True
        self.__logcounter_success = 0 # counts successful publishes!
        self.__logcounter_trigger = 0 # counts triggers!
        self.__LOGFREQUENCY = 10
        self.__have_not_warned_about_connection_fail_yet = True
        self.__have_not_warned_about_force_close_yet = True

    '''
    Triggers the publication of one message to RabbitMQ, if the
    state machine currently allows this.

    The message is fetched from the Queue of unpublished messages.

    So far, whenever the library wants to publish messages, it
    fires as many of these "publish_message" events as messages
    were published (and some extra, to be sure).
    If some of these triggers cannot be acted upon, as the module
    is not in a state where it is allowed to publish, the triggers
    should be fired as soon as the module is in available state
    again.

    # TODO: Find better way to make sure enough publish events are fired.
    Are we sure there is not ever a way to have some messages
    in the unpublished Queue that could be sent, but aren't, because
    no event was fired for them? For example, if an exception occurs
    during publish, and the message was put back - will there ever
    be an event to trigger its publication? I don't think so.

    Interim solution (hack):
    (a) At the moment, for every message that the publisher hands 
        over, I fire two events (rabbitthread).
    (b) During the close-down algorithm, if there is unpublished
        messages, I fire publish events, to make sure they are
        published (thread_shutter).

    '''
    def publish_message(self):
        try:
            return self.__publish_message()
        except Exception as e:
            logwarn(LOGGER, 'Error in feeder.publish_message(): %s: %s', e.__class__.__name__, repr(e))
            raise e

    def __publish_message(self):
        self.__logcounter_trigger += 1

        if self.statemachine.is_NOT_STARTED_YET() or self.statemachine.is_WAITING_TO_BE_AVAILABLE():
            log_every_x_times(LOGGER, self.__logcounter_trigger, self.__LOGFREQUENCY, 'Received early trigger for feeding the rabbit (trigger %i).', self.__logcounter_trigger)
            self.__log_why_cannot_feed_the_rabbit_now()

        elif self.statemachine.is_AVAILABLE() or self.statemachine.is_AVAILABLE_BUT_WANTS_TO_STOP():
            log_every_x_times(LOGGER, self.__logcounter_trigger, self.__LOGFREQUENCY, 'Received trigger for publishing message to RabbitMQ (trigger %i).', self.__logcounter_trigger)
            self.__log_publication_trigger()
            self.__publish_message_to_channel()

        elif self.statemachine.is_PERMANENTLY_UNAVAILABLE() or self.statemachine.is_FORCE_FINISHED():
            log_every_x_times(LOGGER, self.__logcounter_trigger, self.__LOGFREQUENCY, 'Received late trigger for feeding the rabbit (trigger %i).', self.__logcounter_trigger)
            self.__log_why_cannot_feed_the_rabbit_now()

    ''' This method only logs. '''
    def __log_publication_trigger(self):
        if self.__first_publication_trigger:
            logdebug(LOGGER, 'Received first trigger for publishing message to RabbitMQ.')
            self.__first_publication_trigger = False
        logtrace(LOGGER, 'Received trigger for publishing message to RabbitMQ, and module is ready to accept it.')

    ''' This method only logs, depending on the state machine's state.'''
    def __log_why_cannot_feed_the_rabbit_now(self):
        log_every_x_times(LOGGER, self.__logcounter_trigger, self.__LOGFREQUENCY, 'Cannot publish message to RabbitMQ (trigger no. %i).', self.__logcounter_trigger)
        if self.statemachine.is_WAITING_TO_BE_AVAILABLE():
            logdebug(LOGGER, 'Cannot publish message to RabbitMQ yet, as the connection is not ready.')
        elif self.statemachine.is_NOT_STARTED_YET():
            logerror(LOGGER, 'Cannot publish message to RabbitMQ, as the thread is not running yet.')
        elif self.statemachine.is_PERMANENTLY_UNAVAILABLE() or self.statemachine.is_FORCE_FINISHED():
            if self.statemachine.detail_could_not_connect:
                logtrace(LOGGER, 'Could not publish message to RabbitMQ, as the connection failed.')
                if self.__have_not_warned_about_connection_fail_yet:
                    logwarn(LOGGER, 'Could not publish message(s) to RabbitMQ. The connection failed definitively.')
                    self.__have_not_warned_about_connection_fail_yet = False
            elif self.statemachine.get_detail_closed_by_publisher():
                logtrace(LOGGER, 'Cannot publish message to RabbitMQ, as the connection was closed by the user.')
                if self.__have_not_warned_about_force_close_yet:
                    logwarn(LOGGER, 'Could not publish message(s) to RabbitMQ. The sender was closed by the user.')
                    self.__have_not_warned_about_force_close_yet = False
        else:
            if self.thread._channel is None:
                logerror(LOGGER, 'Very unexpected. Could not publish message(s) to RabbitMQ. There is no channel.')

    '''
    Retrieves a message from stack and tries to publish it
    to RabbitMQ.
    In case of failure, it is put back. In case of success,
    it is handed on to the confirm module that is responsible
    for waiting for RabbitMQ's confirmation.

    Note: The publish may cause an error if the Channel was closed.
    A closed Channel should be handled in the on_channel_close()
    callback, but we catch it here in case the clean up was not quick enough.
    '''
    def __publish_message_to_channel(self):

        # Find a message to publish.
        # If no messages left, well, nothing to publish!
        try:
            message = self.__get_message_from_stack()
        except queue.Empty as e:
            logtrace(LOGGER, 'Queue empty. No more messages to be published.')
            return

        # Now try to publish it.
        # If anything goes wrong, you need to put it back to
        # the stack of unpublished messages!
        try:
            success = self.__try_publishing_otherwise_put_back_to_stack(message)
            if success:
                self.__postparations_after_successful_feeding(message)

        # Treat various errors that may occur during publishing:
        except pika.exceptions.ChannelClosed as e:
            logwarn(LOGGER, 'Cannot publish message %i to RabbitMQ because the Channel is closed (%s)', self.__delivery_number+1, repr(e))

        except AttributeError as e:
            if self.thread._channel is None:
                logwarn(LOGGER, 'Cannot publish message %i to RabbitMQ because there is no channel.', self.__delivery_number+1)
            else:
                logwarn(LOGGER, 'Cannot publish message %i to RabbitMQ (unexpected error %s:%s)', self.__delivery_number+1, e.__class__.__name__, repr(e))

        except AssertionError as e:
            logwarn(LOGGER, 'Cannot publish message to RabbitMQ %i because of AssertionError: "%s"', self.__delivery_number+1,e)
            if 'A non-string value was supplied for self.exchange' in repr(e):
                exch = self.thread.get_exchange_name()
                logwarn(LOGGER, 'Exchange was "%s" (type %s)', exch, type(exch))


    '''
    Retrieve an unpublished message from stack.
    Note: May block for up to 2 seconds.

    :return: A message from the stack of unpublished messages.
    :raises: queue.Empty.
    '''
    def __get_message_from_stack(self, seconds=0):
        message = self.thread.get_message_from_unpublished_stack(seconds)
        logtrace(LOGGER, 'Found message to be published. Now left in queue to be published: %i messages.', self.thread.get_num_unpublished())
        return message

    '''
    This tries to publish the message and puts it back into the
    Queue if it failed.

    :param message: Message to be sent.
    :raises: pika.exceptions.ChannelClosed, if the Channel is closed.
    '''
    def __try_publishing_otherwise_put_back_to_stack(self, message):
        try:
            # Getting message info:
            properties = self.nodemanager.get_properties_for_message_publications()
            routing_key, msg_string = rabbitutils.get_routing_key_and_string_message_from_message_if_possible(message)
            routing_key = self.nodemanager.adapt_routing_key_for_untrusted(routing_key)
            
            # Logging
            logtrace(LOGGER, 'Publishing message %i (key %s) (body %s)...', self.__delivery_number+1, routing_key, msg_string) # +1 because it will be incremented after the publish.
            log_every_x_times(LOGGER, self.__logcounter_trigger, self.__LOGFREQUENCY, 'Trying actual publish... (trigger no. %i).', self.__logcounter_trigger)
            logtrace(LOGGER, '(Publish to channel no. %i).', self.thread._channel.channel_number)

            # Actual publish to exchange
            self.thread._channel.basic_publish(
                exchange=self.thread.get_exchange_name(),
                routing_key=routing_key,
                body=msg_string,
                properties=properties,
                mandatory=defaults.RABBIT_MANDATORY_DELIVERY
            )
            return True

        # If anything went wrong, put it back into the stack of
        # unpublished messages before re-raising the exception
        # for further handling:
        except Exception as e:
            success = False
            logwarn(LOGGER, 'Message was not published. Putting back to queue. Reason: %s: "%s"',e.__class__.__name__, repr(e))
            self.thread.put_one_message_into_queue_of_unsent_messages(message)
            logtrace(LOGGER, 'Now (after putting back) left in queue to be published: %i messages.', self.thread.get_num_unpublished())
            raise e

    '''
    If a publish was successful, pass it to the confirmer module
    and in increment delivery_number for the next message.
    '''
    def __postparations_after_successful_feeding(self, msg):

        # Pass the successfully published message and its delivery_number
        # to the confirmer module, to wait for its confirmation.
        # Increase the delivery number for the next message.
        self.thread.put_to_unconfirmed_delivery_tags(self.__delivery_number)
        self.thread.put_to_unconfirmed_messages_dict(self.__delivery_number, msg)
        self.__delivery_number += 1

        # Logging
        self.__logcounter_success += 1
        log_every_x_times(LOGGER, self.__logcounter_success, self.__LOGFREQUENCY, 'Actual publish to channel done (trigger no. %i, publish no. %i).', self.__logcounter_trigger, self.__logcounter_success)
        logtrace(LOGGER, 'Publishing messages %i to RabbitMQ... done.', self.__delivery_number-1)
        if (self.__delivery_number-1 == 1):
            loginfo(LOGGER, 'First message published to RabbitMQ.')
        logdebug(LOGGER, 'Message published (no. %i)', self.__delivery_number-1)

    '''
    Reset the delivery_number for the messages.
    This must be called on a reconnection / channel reopen!
    And may not be called during any other situation!

    The number is not sent along to the RabbitMQ server, but
    the server keeps track of the delivery number
    separately on its side.

    That's why it is important to make sure it is incremented
    and reset exactly the same way (incremented at each successfully
    published message, and reset to one at channel reopen).

    (called by the builder during reconnection / channel reopen).
    '''
    def reset_delivery_number(self):
        self.__delivery_number = 1



import sys
import unittest
import mock
import logging
import copy
import esgfpid.defaults
import esgfpid.utils

# Logging
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

LOGGER_TO_PASS = logging.getLogger('utils')
LOGGER_TO_PASS.addHandler(logging.NullHandler())


class ErrorMessageUtilsTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    #
    # Test getting routing key and string message
    #

    def test_get_message_and_routing_key_string_ok1(self):
        
        # Test variables:
        passed_messages = ['foo', 'bar', 'lorem ipsum lorem ipsum loooorem']

        # Run code to be checked:
        received_message = esgfpid.utils.format_error_message(passed_messages)

        # Check result:
        expected_message = (
            '****************************************\n'+
            '*** foo                              ***\n'+
            '*** bar                              ***\n'+
            '*** lorem ipsum lorem ipsum loooorem ***\n'+
            '****************************************')
        self.assertEqual(expected_message, received_message, 'Wrong message.\nExpected: %s\nReceived: %s' % (expected_message, received_message))


    def test_get_message_and_routing_key_string_ok2(self):
        
        # Test variables:
        passed_messages = ['foo']

        # Run code to be checked:
        received_message = esgfpid.utils.format_error_message(passed_messages)

        # Check result:
        expected_message = (
            '***********\n'+
            '*** foo ***\n'+
            '***********')
        self.assertEqual(expected_message, received_message, 'Wrong message.\nExpected: %s\nReceived: %s' % (expected_message, received_message))




import unittest
import mock
import logging
import datetime
import pika
LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

# Test resources:
from resources.TESTVALUES import *
import resources.TESTVALUES as TESTHELPERS

class ThreadBuilderTestCase(unittest.TestCase):

    def setUp(self):
        LOGGER.info('######## Next test (%s) ##########', __name__)

    def tearDown(self):
        LOGGER.info('#############################')

    def make_builder(self, args=None):

        statemachine = esgfpid.rabbit.asynchronous.thread_statemachine.StateMachine()
        confirmer = esgfpid.rabbit.asynchronous.thread_confirmer.Confirmer()
        nodemanager = TESTHELPERS.get_nodemanager()

        thread = mock.MagicMock()
        thread.ERROR_CODE_CONNECTION_CLOSED_BY_USER=999
        thread.ERROR_TEXT_CONNECTION_FORCE_CLOSED='(forced finish)'
        thread.ERROR_TEXT_CONNECTION_NORMAL_SHUTDOWN='(not reopen)'
        thread.ERROR_TEXT_CONNECTION_PERMANENT_ERROR='(permanent error)'
        returnhandler = mock.MagicMock()
        shutter = mock.MagicMock()

        builder = esgfpid.rabbit.asynchronous.thread_builder.ConnectionBuilder(
            thread,
            statemachine,
            confirmer,
            returnhandler,
            shutter,
            nodemanager)

        return builder

    #
    # first connection
    #

    '''
    Test first connection, normal case.
    '''
    @mock.patch('esgfpid.rabbit.asynchronous.thread_builder.ConnectionBuilder._ConnectionBuilder__please_open_connection')
    def test_first_connection_ok(self, connpatch):

        # Patch
        def make_conn():
            builder.thread._connection = mock.MagicMock()
            builder.thread._connection.ioloop = mock.MagicMock()
            builder._ConnectionBuilder__start_connect_time = datetime.datetime.now()
        connpatch.side_effect = make_conn

        # Preparation:
        builder = self.make_builder()

        # Run code to be tested:
        builder.first_connection()

        # Check result:
        self.assertEqual(connpatch.call_count, 1)
        builder.thread._connection.ioloop.start.assert_called_with()
        builder.thread.continue_gently_closing_if_applicable.assert_called_with()

    '''
    Test first connection, ProbableAuthenticationError.
    The on_connection_error method is called by the library itself,
    and the ioloop is started anyway.
    '''
    @mock.patch('esgfpid.rabbit.asynchronous.thread_builder.ConnectionBuilder._ConnectionBuilder__please_open_connection')
    def test_first_connection_auth_error(self, connpatch):

        # Create a connection object
        mock_connection = mock.MagicMock()
        mock_connection.ioloop = mock.MagicMock()
        mock_connection.ioloop.start = mock.MagicMock()
        side_effect_first_call = pika.exceptions.ProbableAuthenticationError
        side_effect_second_call = None
        mock_connection.ioloop.start.side_effect = (side_effect_first_call, side_effect_second_call)

        # Patched method:
        # This is what should have happened during the patched method...
        def make_conn():
            builder.thread._connection = mock_connection
            builder._ConnectionBuilder__start_connect_time = datetime.datetime.now()
        connpatch.side_effect = make_conn

        # Preparation:
        builder = self.make_builder()

        # Mock the error method
        builder.on_connection_error = mock.MagicMock()

        # Run code to be tested:
        builder.first_connection()

        # Check result:
        self.assertEqual(connpatch.call_count, 1)
        builder.on_connection_error.assert_called()
        mock_connection.ioloop.start.assert_called_with()
        builder.thread.continue_gently_closing_if_applicable.assert_called_with()

    '''
    Test first connection, any other exception.
    I am not sure this can actually happen, or how!
    The on_connection_error method is called by the library itself,
    and the ioloop is started anyway.
    '''
    @mock.patch('esgfpid.rabbit.asynchronous.thread_builder.ConnectionBuilder._ConnectionBuilder__please_open_connection')
    def test_first_connection_any_error(self, connpatch):

        # Create a connection object
        mock_connection = mock.MagicMock()
        mock_connection.ioloop = mock.MagicMock()
        mock_connection.ioloop.start = mock.MagicMock()
        side_effect_first_call = KeyError # any error...
        side_effect_second_call = None
        mock_connection.ioloop.start.side_effect = (side_effect_first_call, side_effect_second_call)

        # Patched method:
        # This is what should have happened during the patched method...
        def make_conn():
            builder.thread._connection = mock_connection
            builder._ConnectionBuilder__start_connect_time = datetime.datetime.now()
        connpatch.side_effect = make_conn

        # Preparation:
        builder = self.make_builder()

        # Mock the error method
        builder.on_connection_error = mock.MagicMock()

        # Run code to be tested:
        builder.first_connection()

        # Check result:
        self.assertEqual(connpatch.call_count, 1)
        builder.on_connection_error.assert_called()
        mock_connection.ioloop.start.assert_called_with()
        builder.thread.continue_gently_closing_if_applicable.assert_called_with()

    #
    # on_connection_open
    #

    '''
    We test the actions when RabbitMQ has opened a connection.
    It triggers opening the channel and stops the publisher from
    blocking.
    '''
    @mock.patch('esgfpid.rabbit.asynchronous.thread_builder.ConnectionBuilder._ConnectionBuilder__please_open_rabbit_channel')
    def test_on_connection_open_ok(self, channelpatch):

        # Patch
        channelpatch.return_value = mock.MagicMock()

        # Preparation:
        builder = self.make_builder()

        # Run code to be tested:
        unused_connection = None
        builder.on_connection_open(unused_connection)

        # Check result:
        self.assertEqual(channelpatch.call_count, 1)
        builder.thread.tell_publisher_to_stop_waiting_for_thread_to_accept_events.assert_called_with()

    #
    # on_channel_open
    #

    '''
    We test the actions when RabbitMQ has opened a channel.
    The channel is set to confirm-deliveries and the callbacks are added.
    '''
    def test_on_channel_open_ok(self):

        # Preparation:
        builder = self.make_builder()
        builder._ConnectionBuilder__start_connect_time = datetime.datetime.now()
        builder.thread.get_num_unpublished.return_value = 0
        builder.statemachine.set_to_waiting_to_be_available()
        testchannel = mock.MagicMock()

        # Run code to be tested:
        builder.on_channel_open(testchannel)

        # Check result:
        self.assertTrue(builder.statemachine.is_AVAILABLE())
        testchannel.confirm_delivery.assert_called()
        testchannel.add_on_return_callback.assert_called()
        testchannel.add_on_close_callback.assert_called()
        builder.thread.get_num_unpublished.assert_called()
        builder.thread.add_event_publish_message.assert_not_called()

    '''
    We test the actions when RabbitMQ has opened a channel,
    and some messages arrived in the meantime (from the publisher).
    The messages are sent.
    '''
    def test_on_channel_open_messages_waiting_ok(self):

        # Preparation:
        builder = self.make_builder()
        builder._ConnectionBuilder__start_connect_time = datetime.datetime.now()
        builder.statemachine.set_to_waiting_to_be_available()
        builder.thread.get_num_unpublished.return_value = 11
        testchannel = mock.MagicMock()

        # Run code to be tested:
        builder.on_channel_open(testchannel)

        # Check result:
        self.assertTrue(builder.statemachine.is_AVAILABLE())
        testchannel.confirm_delivery.assert_called()
        testchannel.add_on_return_callback.assert_called()
        testchannel.add_on_close_callback.assert_called()
        builder.thread.get_num_unpublished.assert_called()
        builder.thread.add_event_publish_message.assert_called()
        builder.thread.add_event_publish_message.call_count >= 11


    '''
    We test the actions when RabbitMQ has opened a channel,
    but in the meantime the library was force-closed.
    No messages are sent, and the library is safety-closed.
    '''
    def test_on_channel_open_force_closed(self):

        # Preparation:
        builder = self.make_builder()
        builder._ConnectionBuilder__start_connect_time = datetime.datetime.now()
        builder.statemachine.set_detail_closed_by_publisher()
        builder.statemachine.set_to_force_finished()
        testchannel = mock.MagicMock()

        # Run code to be tested:
        builder.on_channel_open(testchannel)

        # Check result:
        self.assertTrue(builder.statemachine.is_FORCE_FINISHED())
        self.assertTrue(builder.statemachine.get_detail_closed_by_publisher())
        builder.thread.add_event_publish_message.assert_not_called()
        builder.shutter.safety_finish.assert_called_with('closed before connection was ready. reclosing.')

    '''
    We test the actions when RabbitMQ has opened a channel,
    but in the meantime the library was force-closed.
    No messages are sent, although some were waiting, and
    the library is safety-closed.
    '''
    def test_on_channel_open_force_closed_pending_messages(self):

        # Preparation:
        builder = self.make_builder()
        builder._ConnectionBuilder__start_connect_time = datetime.datetime.now()
        builder.statemachine.set_detail_closed_by_publisher()
        builder.statemachine.set_to_force_finished()
        builder.thread.get_num_unpublished.return_value = 11
        testchannel = mock.MagicMock()

        # Run code to be tested:
        builder.on_channel_open(testchannel)

        # Check result:
        self.assertTrue(builder.statemachine.is_PERMANENTLY_UNAVAILABLE())
        builder.thread.add_event_publish_message.assert_not_called()
        builder.shutter.safety_finish.assert_called_with('closed before connection was ready. reclosing.')

    '''
    We test the actions when RabbitMQ has opened a channel,
    but in the meantime the library was gently closed.
    '''
    def test_on_channel_open_gently_closed(self):

        # Preparation:
        builder = self.make_builder()
        builder._ConnectionBuilder__start_connect_time = datetime.datetime.now()
        builder.statemachine.set_to_wanting_to_stop()
        builder.thread.get_num_unpublished.return_value = 0
        testchannel = mock.MagicMock()

        # Run code to be tested:
        builder.on_channel_open(testchannel)

        # Check result:
        self.assertFalse(builder.statemachine.is_AVAILABLE())
        self.assertTrue(builder.statemachine.is_AVAILABLE_BUT_WANTS_TO_STOP())
        builder.thread.add_event_publish_message.assert_not_called()
        builder.shutter.safety_finish.assert_not_called()

    '''
    We test the actions when RabbitMQ has opened a channel,
    but in the meantime the library was gently closed.
    The pending messages are published.
    '''
    def test_on_channel_open_gently_closed_pending(self):

        # Preparation:
        builder = self.make_builder()
        builder._ConnectionBuilder__start_connect_time = datetime.datetime.now()
        builder.statemachine.set_to_wanting_to_stop()
        builder.thread.get_num_unpublished.return_value = 11
        testchannel = mock.MagicMock()

        # Run code to be tested:
        builder.on_channel_open(testchannel)

        # Check result:
        self.assertTrue(builder.statemachine.is_AVAILABLE_BUT_WANTS_TO_STOP())
        testchannel.confirm_delivery.assert_called()
        testchannel.add_on_return_callback.assert_called()
        testchannel.add_on_close_callback.assert_called()
        builder.thread.get_num_unpublished.assert_called()
        builder.thread.add_event_publish_message.assert_called()
        builder.thread.add_event_publish_message.call_count >= 11
        builder.shutter.safety_finish.assert_not_called()

    #
    # on_connection_error
    #

    '''
    Testing behaviour on connection errors, if only one RabbitMQ
    instances has been tried, and more are available.
    It tries to connect to the next.
    '''
    @mock.patch('esgfpid.rabbit.asynchronous.thread_builder.ConnectionBuilder._ConnectionBuilder__please_open_connection')
    def test_on_connection_error_has_more(self, connpatch):

        # Preparation:
        builder = self.make_builder()
        builder.thread.get_num_unpublished.return_value = 11
        builder._ConnectionBuilder__start_connect_time = datetime.datetime.now()
        builder.statemachine.set_to_waiting_to_be_available()

        # Mock the connection
        mock_connection = mock.MagicMock()
        def side_effect_call_later(wait_seconds, callback):
            callback()
        mock_connection.ioloop.call_later.side_effect = side_effect_call_later

        # Run code to be tested:
        builder.on_connection_error(mock_connection, 'foo message')

        # Check result:
        # Reconnect was called:
        wait_seconds = 0
        mock_connection.ioloop.call_later.assert_called_with(wait_seconds, builder.reconnect)
        # This was called inside reconnect:
        builder.thread._connection.ioloop.stop.assert_called()
        connpatch.assert_called()
        self.assertEqual(connpatch.call_count, 1)
        # As the connection is never built, no messages are sent, it is not set to available:
        builder.thread.add_event_publish_message.assert_not_called()
        self.assertFalse(builder.statemachine.is_AVAILABLE())

    '''
    Testing behaviour on connection errors, if there was a
    force-finish in the meantime.
    '''
    @mock.patch('esgfpid.rabbit.asynchronous.thread_builder.ConnectionBuilder._ConnectionBuilder__please_open_connection')
    def test_on_connection_error_force_finish(self, connpatch):

        # Preparation:
        builder = self.make_builder()
        builder.thread.get_num_unpublished.return_value = 11
        builder._ConnectionBuilder__start_connect_time = datetime.datetime.now()
        builder.statemachine.set_to_force_finished()

        # Mock the connection
        mock_connection = mock.MagicMock()
        def side_effect_call_later(wait_seconds, callback):
            callback()
        mock_connection.ioloop.call_later.side_effect = side_effect_call_later

        # Run code to be tested:
        with self.assertRaises(esgfpid.rabbit.exceptions.PIDServerException) as e:
            builder.on_connection_error(mock_connection, 'foo message')

        # Check result:
        # Reconnect was not called:
        mock_connection.ioloop.call_later.assert_not_called()
        connpatch.assert_not_called()
        # As the connection is never built, no messages are sent, it is not set to available:
        builder.thread.add_event_publish_message.assert_not_called()
        self.assertFalse(builder.statemachine.is_AVAILABLE())
        self.assertTrue(builder.statemachine.is_FORCE_FINISHED())

    '''
    Testing behaviour on connection errors, if all RabbitMQ
    instances have been tried once.
    It is reset, all RabbitMQ instances are tried again after
    a waiting period.
    '''
    @mock.patch('esgfpid.rabbit.asynchronous.thread_builder.ConnectionBuilder._ConnectionBuilder__please_open_connection')
    def test_on_connection_error_none_left_reset(self, connpatch):

        # Preparation:
        builder = self.make_builder()
        builder.thread.get_num_unpublished.return_value = 11
        builder._ConnectionBuilder__start_connect_time = datetime.datetime.now()
        builder._ConnectionBuilder__node_manager.set_next_host() # otherwise, we get errors, if no current host is available at all.
        builder._ConnectionBuilder__node_manager._NodeManager__trusted_nodes = {}
        builder.statemachine.set_to_waiting_to_be_available()

        # Mock the connection
        mock_connection = mock.MagicMock()
        def side_effect_call_later(wait_seconds, callback):
            callback()
        mock_connection.ioloop.call_later.side_effect = side_effect_call_later

        # Run code to be tested:
        builder.on_connection_error(mock_connection, 'foo message')

        # Check result:
        # Reconnect was called:
        wait_seconds = esgfpid.defaults.RABBIT_RECONNECTION_SECONDS
        mock_connection.ioloop.call_later.assert_called_with(wait_seconds, builder.reconnect)
        # This was called inside reconnect:
        builder.thread._connection.ioloop.stop.assert_called()
        connpatch.assert_called()
        self.assertEqual(connpatch.call_count, 1)
        # As the connection is never built, no messages are sent, it is not set to available:
        builder.thread.add_event_publish_message.assert_not_called()
        self.assertFalse(builder.statemachine.is_AVAILABLE())

    '''
    Testing behaviour on connection errors, if all RabbitMQ
    instances have been tried several times...
    It is set to permanently unavailable, no more reconnections.
    '''
    def test_on_connection_error_none_left_no_resetting(self):

        # Preparation:
        builder = self.make_builder()
        builder.thread.get_num_unpublished.return_value = 11
        builder._ConnectionBuilder__start_connect_time = datetime.datetime.now()
        builder._ConnectionBuilder__node_manager.set_next_host() # otherwise, we get errors, if no current host is available at all.
        builder._ConnectionBuilder__node_manager._NodeManager__trusted_nodes = {}
        builder._ConnectionBuilder__reconnect_counter = esgfpid.defaults.RABBIT_RECONNECTION_MAX_TRIES
        builder.statemachine.set_to_waiting_to_be_available()

        # Run code to be tested:
        mock_connection = mock.MagicMock()
        with self.assertRaises(esgfpid.rabbit.exceptions.PIDServerException) as e:
            builder.on_connection_error(mock_connection, 'foo message')

        # Check result:
        self.assertIn('Permanently failed to connect to RabbitMQ.', str(e.exception))
        self.assertIn('Tried all hosts 3 times.', str(e.exception))
        self.assertIn('Giving up. No PID requests will be sent.', str(e.exception))
        mock_connection.ioloop.call_later.assert_not_called()
        builder.thread.add_event_publish_message.assert_not_called()
        builder.thread.add_event_publish_message.call_count >= 0
        self.assertTrue(builder.statemachine.is_PERMANENTLY_UNAVAILABLE())
        self.assertTrue(builder.statemachine.detail_could_not_connect)

    def test_on_connection_error_force_finished_2(self):

        # Preparation:
        builder = self.make_builder()
        builder.thread.get_num_unpublished.return_value = 11
        builder._ConnectionBuilder__start_connect_time = datetime.datetime.now()
        builder.statemachine.set_to_waiting_to_be_available()

        # Make sure the is_FORCE_FINISH returns False the first time and true the second time...
        builder.statemachine.is_FORCE_FINISHED = mock.MagicMock()
        builder.statemachine.is_FORCE_FINISHED.side_effect = [False, True, True, True]

        # Mock the connection
        mock_connection = mock.MagicMock()

        # Run code to be tested:
        with self.assertRaises(esgfpid.rabbit.exceptions.PIDServerException) as e:
            builder.on_connection_error(mock_connection, 'foo message')

        

        # Check result:
        self.assertIn('until received a force-finish. Giving up', str(e.exception))
        builder.statemachine.is_FORCE_FINISHED.assert_called()
        self.assertEqual(builder.statemachine.is_FORCE_FINISHED.call_count, 4)
        # Reconnect was called:
        mock_connection.ioloop.call_later.assert_not_called()
        # As the connection is never built, no messages are sent, it is not set to available:
        builder.thread.add_event_publish_message.assert_not_called()
        self.assertFalse(builder.statemachine.is_AVAILABLE())

    #
    # on_channel_close
    #

    '''
    The channel was closed because the publisher closed
    the library.
    Requires no action. User-close will call the on_connection_close
    callback, so everything is handled there.
    '''
    def test_on_channel_closed_by_publisher(self):

        # Preparation:
        builder = self.make_builder()
        builder.statemachine.set_to_permanently_unavailable()
        builder.statemachine.set_detail_closed_by_publisher()

        # Run code to be tested:
        testchannel = None
        exception = None
        builder.on_channel_closed(testchannel, exception)

        # Check result
        builder.thread._connection.close.assert_not_called()

    '''
    The channel was closed unexpectedly.
    This can be caused by any event from the RabbitMQ.
    So we trigger a reconnection (which is triggered by
    closing the connection).
    '''
    def test_on_channel_closed_unexpected_ok(self):

        # Preparation:
        builder = self.make_builder()
        builder.statemachine.set_to_available()
        builder.thread._connection = mock.MagicMock()

        # Run code to be tested:
        testchannel = None
        exception = None
        builder.on_channel_closed(testchannel, exception)

        # Check result
        builder.thread._connection.close.assert_called_with()

    '''
    The channel was closed because the desired exchange did
    not exist.
    '''
    def test_on_channel_no_exchange(self):

        # Preparation:
        builder = self.make_builder()
        builder.statemachine.set_to_available()
        builder.thread._connection = mock.MagicMock()

        # Run code to be tested:
        testchannel = None
        exception = pika.exceptions.ChannelClosed(404, 'foo')
        builder.on_channel_closed(testchannel, exception)

        # Check result
        builder.thread._connection.close.assert_not_called()
        builder.thread.change_exchange_name.assert_called_with(esgfpid.defaults.RABBIT_FALLBACK_EXCHANGE_NAME)
        builder.thread._connection.channel.assert_called()

    '''
    The channel was closed because the fallback exchange did
    not exist.
    We close the connection. This triggers reconnection to
    another host (not in this test, as there is not RabbitMQ
    to call a callback.)
    '''
    def test_on_channel_no_fallback_exchange(self):

        # Preparation:
        builder = self.make_builder()
        builder.statemachine.set_to_available()
        builder.thread._connection = mock.MagicMock()

        # Run code to be tested:
        testchannel = None
        exception = pika.exceptions.ChannelClosed(404, "NOT_FOUND - no exchange 'FALLBACK'")
        builder.on_channel_closed(testchannel, exception)

        # Check result
        builder.thread._connection.close.assert_called()

    #
    # on_connection_close
    #

    def test_on_connection_closed_gentle_ok(self):

        # Preparation:
        builder = self.make_builder()
        builder.statemachine.set_to_available()

        # Run code to be tested:
        testconnection = None
        exception = pika.exceptions.ConnectionClosed(999, 'foo(not reopen)foo')
        builder.on_connection_closed(testconnection, exception)

        # Check result
        builder.thread._connection.ioloop.stop.assert_called_with()
        self.assertTrue(builder.statemachine.is_PERMANENTLY_UNAVAILABLE())

    def test_on_connection_closed_forced_ok(self):

        # Preparation:
        builder = self.make_builder()
        builder.statemachine.set_to_available()
        builder.statemachine.closed_by_publisher = False

        # Run code to be tested:
        testconnection = None
        exception = pika.exceptions.ConnectionClosed(999, 'foo(forced finish)foo')
        builder.on_connection_closed(testconnection, exception)

        # Check result
        builder.thread._connection.ioloop.stop.assert_called_with()
        self.assertTrue(builder.statemachine.is_PERMANENTLY_UNAVAILABLE())


    def test_on_connection_closed_unexpected_ok(self):

        # Preparation:
        builder = self.make_builder()
        builder.statemachine.set_to_available()
        builder.statemachine.closed_by_publisher = False
        builder._ConnectionBuilder__start_connect_time = datetime.datetime.now()

        # Run code to be tested:
        exception = pika.exceptions.ConnectionClosed(111, '(foo)')
        builder.on_connection_closed(builder.thread._connection, exception)

        # Check result
        builder.thread._connection.ioloop.stop.assert_not_called()
        self.assertTrue(builder.statemachine.is_WAITING_TO_BE_AVAILABLE())
        builder.thread._connection.ioloop.call_later.assert_called_with(0, builder.reconnect)

    def test_on_connection_closed_permanent_error_ok(self):

        # Preparation:
        builder = self.make_builder()
        builder.statemachine.set_to_available()
        builder.statemachine.closed_by_publisher = False

        # Run code to be tested:
        testconnection = None
        exception = pika.exceptions.ConnectionClosed(999, 'foo(permanent error)foo')
        builder.on_connection_closed(testconnection, exception)

        # Check result
        builder.thread._connection.ioloop.stop.assert_called_with()
        self.assertTrue(builder.statemachine.is_PERMANENTLY_UNAVAILABLE())

    #
    # reconnect
    #

    @mock.patch('esgfpid.rabbit.asynchronous.thread_builder.ConnectionBuilder._ConnectionBuilder__please_open_connection')
    def test_reconnect_no_rescued_messages(self, connpatch):

        # Preparation:
        builder = self.make_builder()

        # Run code to be tested:
        builder.reconnect()

        # Check result:
        builder.thread._connection.ioloop.stop.assert_called_with()
        builder.thread.reset_delivery_number.assert_called_with()
        builder.thread.reset_unconfirmed_messages_and_delivery_tags.assert_called_with()
        builder.thread.send_many_messages.assert_not_called()
        connpatch.assert_called()


    @mock.patch('esgfpid.rabbit.asynchronous.thread_builder.ConnectionBuilder._ConnectionBuilder__please_open_connection')
    def test_reconnect_some_rescued_messages(self, connpatch):

        # Preparation:
        builder = self.make_builder()
        builder.thread.get_num_unpublished.return_value = 11
        builder.thread.get_unconfirmed_messages_as_list_copy_during_lifetime = mock.MagicMock()
        builder.thread.get_unconfirmed_messages_as_list_copy_during_lifetime.return_value = ['foo','bar','baz']

        # Run code to be tested:
        builder.reconnect()

        # Check result:
        builder.thread._connection.ioloop.stop.assert_called_with()
        builder.thread.reset_delivery_number.assert_called_with()
        builder.thread.reset_unconfirmed_messages_and_delivery_tags.assert_called_with()
        builder.thread.send_many_messages.assert_called()
        connpatch.assert_called()





# Making this available directly via esgfpid.check_pid_queue_availability
# instead of esgfpid.check.check_pid_queue_availability
from .check import check_pid_queue_availability

# Making this available directly via esgfpid.make_handle_from_drsid_and_versionnumber
# instead of esgfpid.utils.make_handle_from_drsid_and_versionnumber
from .utils import make_handle_from_drsid_and_versionnumber

# Making this available directly via esgfpid.Connector
# instead of esgfpid.connector.Connector
from .connector import Connector




import os

def get_absolute_path(path, file_path_string):
    if os.path.isabs(path):
        return path

    elif path.startswith(os.path.curdir):
        path = path.lstrip(os.path.curdir)
        pathlist = path.split(os.path.sep)
        thisdir = get_this_directory(file_path_string, as_list=True)
        newdir = thisdir + list(pathlist)
        return os.path.sep+os.path.join(*newdir)

    else:
        raise ValueError('Path is neither absolute nor relative.')

def get_this_directory(file_path_string, as_list=False):
    this_directory_string = os.path.split(os.path.realpath(file_path_string))[0]

    if as_list:
        this_directory_list = this_directory_string.split(os.path.sep)
        return this_directory_list
    else:
        return this_directory_string

def get_super_directory(file_path_string, as_list=False):
    this_directory_list = get_this_directory(file_path_string, as_list=True)
    super_directory_list = this_directory_list[0:len(this_directory_list)-1]

    if as_list:
        return super_directory_list
    else:
        super_directory_string = os.path.join(*super_directory_list)
        return os.path.sep+super_directory_string


def get_neighbour_directory(file_path_string, dirname):
    super_directory_list = get_super_directory(file_path_string, as_list=True)
    super_directory_list.append(dirname)
    neighbour_directory_string = os.path.join(*super_directory_list)
    return os.path.sep+neighbour_directory_string

def get_super_neighbour_directory(file_path_string, dirname):
    this_directory = get_this_directory(file_path_string)
    super_super_directory_list = get_super_directory(this_directory, as_list=True)
    super_super_directory_list.append(dirname)
    target_directory_string = os.path.join(*super_super_directory_list)
    return os.path.sep+target_directory_string


import copy

def format_error_message(msg_list):
    max_len = _get_max_length(msg_list)
    msg_list_updated = _fill_up_strings(msg_list, max_len)
    msg_list_updated = _append_line_enclosers(msg_list_updated)
    error_message_string = ('\n'.join(msg_list_updated))
    max_len_new = _get_max_length(msg_list_updated)
    start_end_line = '*' * max_len_new
    error_message_string = start_end_line +'\n'+ error_message_string +'\n'+ start_end_line
    return error_message_string

def _get_max_length(msg_list):
    max_len = 0
    for msg in msg_list:
        if len(msg) > max_len:
            max_len = len(msg)
    return max_len

def _fill_up_strings(msg_list, max_len):
    msg_list_updated = copy.copy(msg_list)
    for i in range(len(msg_list)):
        current_length = len(msg_list[i])
        need_to_fill = max_len - current_length
        msg_list_updated[i] = msg_list[i] + (' '*need_to_fill)
    return msg_list_updated

def _append_line_enclosers(msg_list, encl='***'):
    msg_list_updated = copy.copy(msg_list)
    for i in range(len(msg_list)):
        msg_list_updated[i] = encl + ' ' + msg_list[i] + ' '+ encl
    return msg_list_updated


import uuid
from .argsutils import check_presence_of_mandatory_args

def make_handle_from_drsid_and_versionnumber(**args):
    check_presence_of_mandatory_args(args, ['drs_id','version_number','prefix'])
    suffix = make_suffix_from_drsid_and_versionnumber(drs_id=args['drs_id'], version_number=args['version_number'])
    return _suffix_to_handle(args['prefix'], suffix)

def _suffix_to_handle(prefix, suffix):
    return 'hdl:'+prefix+'/'+suffix

def concatenate_drs_and_versionnumber(drs_id, version_number):
    return drs_id+'.v'+str(version_number)

def make_suffix_from_drsid_and_versionnumber(**args):
    check_presence_of_mandatory_args(args, ['drs_id','version_number'])
    hash_basis = concatenate_drs_and_versionnumber(args['drs_id'], args['version_number'])
    return _make_uuid_from_basis(hash_basis)

def _make_uuid_from_basis(hash_basis):
    ds_uuid = uuid.uuid3(uuid.NAMESPACE_URL, hash_basis) # Using uuid3, as this is easy to use also with Java
    return str(ds_uuid)
    # Which NAMESPACE (1st arg) we use does not matter, as uuids only have to be unique inside ESGF, not globally.
    # It just has to stay the same all the time! Otherwise we will create
    # different handles for the same dataset.

def make_handle_from_list_of_strings(sorted_list_of_strings, prefix, addition=None):
    concatenated = ''.join(sorted_list_of_strings)

    # We add a string to the string to make sure the result is not
    # the same for a shopping cart with only one dataset, and for
    # the dataset itself.
    if addition is not None:
        concatenated = str(addition)+concatenated
    
    suffix = _make_uuid_from_basis(concatenated)
    return _suffix_to_handle(prefix, suffix)

def make_sorted_lowercase_list_without_hdl(strings):
    # Make lowercase
    # Remove "hdl:"
    # Sort list
    newlist = []
    for string in strings:
        string = string.lower()
        string = string.replace('hdl:', '', 1)
        newlist.append(string)
    newlist.sort()
    return newlist


import sys
if sys.version[0] == '2':
    import Queue as queue
else:
    import queue as queue
import threading
import pika
import time
import copy
import datetime
import logging
import esgfpid.utils
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn, log_every_x_times
from .thread_returnhandler import UnacceptedMessagesHandler
from .thread_statemachine import StateMachine
from .thread_builder import ConnectionBuilder
from .thread_feeder import RabbitFeeder
from .thread_shutter import ShutDowner
from .thread_confirmer import Confirmer
from .exceptions import OperationNotAllowed

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

'''
Statemachine is WRITTEN by thread, READ by main.
All other thread_ modules are only WRITTEN by thread.

Main can only influence the thread via run() and via add_event(), and by passing
messages into the thread-safe queue.

How can main get information from the thread?
'''


'''
The class RabbitThread is the parallel thread that is
responsible for the asynchronous communication with the
RabbitMQ server.

It has several submodules that do the actual work - connecting
to RabbitMQ, publishing messages, receiving confirms, handling
connection problems, ... (Those submodules are named "thread_...").

RabbitThread implements its own thread, which, after connecting to
RabbitMQ, basically stays in a blocking loop (ioloop) waiting
for events.

These events can come from the RabbitMQ (via defined callback
methods) or from the main thread. Only the connector object
(AsynchronousRabbitConnector) has a reference to the thread and
passes events to it.

The RabbitThread provides some methods to be called by the main
thread, i.e. by the AsynchronousRabbitConnector, which use pika's
SelectConnection.ioloop.call_later() to pass the event to the thread's
blocking loop.

Only the methods
 * run()
 * add_event_publish_message()
 * add_event_gently_finish()
 * add_event_force_finish()
should be called from the AsynchronousRabbitConnector.

All other methods are thread-UNsafe and to be used by the
RabbitThread's submodules only. They are only public to
enable the submodules to communicate among each other
without needed references to each other.

'''
class RabbitThread(threading.Thread):

    def __init__(self, statemachine, msg_queue, facade, node_manager):
        threading.Thread.__init__(self)

        '''
        The rabbit module used by the library to publish modules
        to this thread. In some situations the thread itself needs
        to be able to publish messages the same way.
        Type: esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector.
        '''
        self.__facade = facade

        '''
        State-machine.
        Shared with the main thread!
        Not thread-safe.
        '''
        self.__statemachine = statemachine

        '''
        Synchronization event which makes sure the main thread
        waits for the rabbit thread to finish shutting down, without
        having to use time.sleep().
        Shared with the main thread!
        '''
        self.__gently_finish_ready = threading.Event()

        '''
        Synchronization event which makes sure the main thread
        waits for the rabbit thread to be ready for events before
        sending events, without having to use time.sleep().
        Shared with the main thread!
        '''
        self.__connection_is_set = threading.Event()
        
        '''
        Thread-safe Queue that will contain the unpublished messages.
        The main thread will put messages into it, and the rabbit thread 
        will retrieve and publish them.
        Shared with the main thread!
        '''
        self.__unpublished_messages_queue = msg_queue

        # These are only used by the rabbit thread:
        '''
        If the messages should not be published to the exchange that
        was passed from the publisher in config, but to a fallback 
        solution, this will be set:
        '''
        self.__fallback_exchange = None
        
        '''
        An object of type "pika.SelectConnection".
        It contains the connection to the RabbitMQ.
        It provides a channel to do the publications and
        an ioloop that blocks and waits for events.

        The connection is initialized by the builder submodule.
        Not used by feeder.
        '''
        self._connection = None

        '''
        An object of type "pika.channel.Channel".

        Used by the feeder for basic_publish, for logging
        its channel number, and for checking whether it
        exists (before publishing).
        '''
        self._channel = None

        # Submodules that do the actual work:
        self.__nodemanager = node_manager
        self.__confirmer = Confirmer()
        self.__returnhandler = UnacceptedMessagesHandler(self)
        self.__feeder = RabbitFeeder(self, self.__statemachine, self.__nodemanager)
        self.__shutter = ShutDowner(self, self.__statemachine)

        '''
        Needed to trigger the connection in run()
        '''
        self.__builder = ConnectionBuilder(self, self.__statemachine, self.__confirmer, self.__returnhandler, self.__shutter, node_manager)


        '''
        This determines how many seconds the client will wait until a message
        is published to the server, after having handed it over to the "connection".
        Obviously, zero is highly recommended.

        The fact that it is possible to increase it requires some precautions.
        If messages can be scheduled to be sent to server, but not actually sent
        yet makes it difficult to loop over the Queue of unsent messages, using
        "while" and "queue.Empty" to break it.
        (It will loop for too long and then have to call the publish method many
        many times with no more messages to publish.)

        If you ever decide to while-loop and break on queue.Empty, just make sure
        the publish interval is zero.
        '''
        self.__PUBLISH_INTERVAL_SECONDS = 0

        # Error codes
        self.ERROR_CODE_CONNECTION_CLOSED_BY_USER=999
        self.ERROR_TEXT_CONNECTION_FORCE_CLOSED='(forced finish)'
        self.ERROR_TEXT_CONNECTION_NORMAL_SHUTDOWN='(not reopen)'
        self.ERROR_TEXT_CONNECTION_PERMANENT_ERROR='(permanent error)'

    #
    # Methods called from the main thread
    #

    '''
    The run() method to implement threading.Thread.

    run() is called once in the beginning.
    Everything that is called inside run() is executed by
    the rabbit thread.
    '''
    def run(self):
        logdebug(LOGGER, 'Starting thread...')
        self.__builder.first_connection()

    '''
    At close down because of permanent errors.
    We need to make sure the main thread does not wait for 
    any event.
    '''
    def unblock_events(self):
        self.__gently_finish_ready.set()
        self.__connection_is_set.set()

    def add_event_publish_message(self):
        logdebug(LOGGER, 'Asking rabbit thread to publish a message...')
        self.__add_event(self.__feeder.publish_message)
        self.__add_event(self.__feeder.publish_message) # Send two...
        logdebug(LOGGER, '(Trigger sent.)')

    def add_event_force_finish(self):
        logdebug(LOGGER, 'Asking rabbit thread to finish quickly...')
        self.__add_event(self.__shutter.force_finish)

    def add_event_gently_finish(self):
        logdebug(LOGGER, 'Asking rabbit thread to finish...')
        self.__add_event(self.__shutter.finish_gently)
        self.__wait_for_thread_to_finish_gently()
        # This waiting is necessary, because after finishing gently
        # the main thread needs to join the rabbit thread, which
        # only makes sense if finishing gently is done.
        # If we separated the finish_gently command and the join()
        # command, the main thread could do other stuff in the mean-
        # time, but we'd have to trust the user to call both, to
        # avoid missing the join.

    '''
    Allows the main thread to retrieve all unconfirmed messages.
    As the method is not thread safe, it can be called only
    after the thread is joined.

    It makes a new copy, so the returned list may be modified
    without causing any harm.

    Note: If the method needs to be called inside the running
    thread, it can be called on the confirmer module directly,
    not using this facade.

    :raises: OperationNotAllowed: If the thread is still alive.
    :return: A list containing all unconfirmed messages.
    '''
    def get_unconfirmed_messages_as_list_copy(self):
        if not self.is_alive():
            return self.__confirmer.get_unconfirmed_messages_as_list_copy()
        else:
            raise OperationNotAllowed('thread has not finished','retrieving unconfirmed messages')

    '''
    Allows the main thread to retrieve all nacked messages.
    As the method is not thread safe, it can be called only
    after the thread is joined.

    It makes a new copy, so the returned list may be modified
    without causing any harm.

    Note: If the method needs to be called inside the running
    thread, it can be called on the confirmer module directly,
    not using this facade.

    :raises: OperationNotAllowed: If the thread is still alive.
    :return: A list containing all nacked messages.
    '''
    def get_nacked_messages_as_list(self):
        if not self.is_alive():
            return self.__confirmer.get_copy_of_nacked()
        else:
            raise OperationNotAllowed('thread has not finished','retrieving nacked messages')
 

    #
    # Helpers
    #

    '''
    This is how to add an event to the rabbit thread
    from the main thread. 
    '''
    def __add_event(self, event):
        if self._connection is not None:
            self._connection.ioloop.call_later(self.__PUBLISH_INTERVAL_SECONDS, event)
        else:
            # If the main thread wants to add an event so quickly after starting the
            # thread that not even the connection object is listening for events yet,
            # we need to force it to wait.
            # Event listening is the first thing that happens when a thread is started,
            # but e.g. for shopping carts, the main thread just sends one message and
            # then wants to close again.
            # In that case, the thread cannot even receive the close event, as it is
            # not started yet.
            logdebug(LOGGER, 'Main thread wants to add event to thread that is not ready to receive events yet. Blocking and waiting.')
            self.__wait_for_thread_to_accept_events()
            logdebug(LOGGER, 'Thread declared itself ready to receive events.')
            self._connection.ioloop.call_later(self.__PUBLISH_INTERVAL_SECONDS, event)
            logerror(LOGGER, 'Added event after having waited for thread to open.')



    '''
    Force the main thread to wait for the rabbit thread 
    while it tries to finish gently.
    Called from "add_event_gently_finish()".
    Executed by the main thread.
    BLOCKS THE MAIN THREAD (this is its purpose!)
    '''
    def __wait_for_thread_to_finish_gently(self):
        # This is executed by the main thread and thus should be placed in the
        # main thread's method, but it is so important not to forget to wait,
        # that it is better to have it boung to the thread's method "add_event_gently_finish()"
        logdebug(LOGGER, 'Now waiting for gentle close-down of RabbitMQ connection...')
        self.__gently_finish_ready.wait()
        logdebug(LOGGER, 'Finished waiting for gentle close-down of RabbitMQ connection.')

    def __wait_for_thread_to_accept_events(self):
        logdebug(LOGGER, 'Now waiting for the connection before I can even close RabbitMQ connection...')
        self.__connection_is_set.wait()

    def tell_publisher_to_stop_waiting_for_thread_to_accept_events(self):
        self.__connection_is_set.set()
        logdebug(LOGGER, 'Finished waiting for thread to start.')
 
    #
    # Methods called from inside the thread
    # These are to be called only by the submodules!
    # As they are not necessarily thread-safe, and
    # may change the functioning of the module!
    #

    '''
    Fire the event (threading.Event)!

    Called by shutter, to tell when the main thread should
    not wait any longer for the gently-finish command.
    (Previously, this was done by using time.sleep() in
    the main thread, but that causes problems, as it messes
    up the proper and punctual receival of RabbitMQ confirms.
    '''
    def tell_publisher_to_stop_waiting_for_gentle_finish(self):
        self.__gently_finish_ready.set()

    '''
    Called by confirmer, once the connection is there, to see if any
    messages have arrived in the mean time to be published.

    Called by feeder, for logging how many messages are left after
    a publish action.

    Called by shutter, to check if all messages were published.
    '''
    def get_num_unpublished(self):
        return self.__unpublished_messages_queue.qsize()

    ''' Called by shutter, to check if all messages were confirmed. '''
    def get_num_unconfirmed(self):
        return self.__confirmer.get_num_unconfirmed()

    ''' Called by feeder, to publish a message. May raise queue.Empty. '''
    def get_message_from_unpublished_stack(self, seconds):
        return self.__unpublished_messages_queue.get(block=True, timeout=seconds) # can raise queue.Empty

    ''' Called by feeder, to put a message back that was not successfully published. '''
    def put_one_message_into_queue_of_unsent_messages(self, message):
        return self.__unpublished_messages_queue.put(message, block=False)

    ''' Called by builder, to republish messages after a reconnection. '''
    def send_many_messages(self, messages):
        return self.__facade.send_many_messages_to_queue(messages)

    '''Called by returnhandle, to republish a message that was not accepted.'''
    def send_a_message(self, message):
        return self.__facade.send_message_to_queue(message)

    '''Called by feeder, to notify confirmer about which message it needs to get confirmed. '''
    def put_to_unconfirmed_delivery_tags(self, delivery_tag):
        return self.__confirmer.put_to_unconfirmed_delivery_tags(delivery_tag)

    '''Called by feeder, to notify confirmer about which message it needs to get confirmed. '''
    def put_to_unconfirmed_messages_dict(self, delivery_tag, message):
        return self.__confirmer.put_to_unconfirmed_messages_dict(delivery_tag, message)

    ''' Called by builder, to prepare message republication after reconnect/channel reopen. '''
    def reset_unconfirmed_messages_and_delivery_tags(self):
        return self.__confirmer.reset_unconfirmed_messages_and_delivery_tags()

    ''' Called by builder, to prepare message republication after reconnect/channel reopen. '''
    def get_unconfirmed_messages_as_list_copy_during_lifetime(self):
        return self.__confirmer.get_unconfirmed_messages_as_list_copy()

    ''' Called by builder, to prepare message republication after reconnect/channel reopen. '''
    def reset_delivery_number(self):
        return self.__feeder.reset_delivery_number()

    ''' Called by feeder. Called by builder, only for logging. '''
    def get_exchange_name(self):
        if self.__fallback_exchange is not None:
            return self.__fallback_exchange
        else:
            return self.__nodemanager.get_exchange_name()

    ''' Called by builder, in case the old exchange caused an error.'''
    def change_exchange_name(self, new_name):
        self.__fallback_exchange = new_name

    ''' Called by builder, in case the old exchange caused an error.'''
    def reset_exchange_name(self):
        self.__fallback_exchange = None

    ''' Called by shutter, in case a connectio is already closing/closed... '''
    def make_permanently_closed_by_user(self):
        return self.__builder.make_permanently_closed_by_user()

    ''' Called by builder any time a new ioloop starts to listen.'''
    def continue_gently_closing_if_applicable(self):
        return self.__shutter.continue_gently_closing_if_applicable()







from .miscutils import *
from .timeutils import *
from .logutils import *
from .argsutils import *
from .error_message_utils import *
from .handleutils import * # This needs argsutils
from .routingkeys import *




import esgfpid.rabbit.rabbit
import esgfpid.check
import logging
import time
import datetime

TESTVALUES_REAL = dict(
    url_messaging_service='handle-esgf.dkrz.de',
    messaging_exchange='rabbitsender_integration_tests',
    exchange_no_queue='rabbitsender_integration_tests_no_queue',
    exchange_name='rabbitsender_integration_tests', # TODO JUST ONE NAME!
    rabbit_username='sendertests',
    rabbit_password='testmebaby',
    routing_key='test_routing_key',
)

#logging.basicConfig(level=logging.DEBUG)
logdate = datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S')
FORMAT = "%(asctime)-15s %(levelname)s:%(name)s:%(message)s"  
logging.basicConfig(level=logging.DEBUG, filename='test_log_'+logdate+'.log', filemode='w', format=FORMAT)


args = dict(
    messaging_service_username = TESTVALUES_REAL['rabbit_username'],
    messaging_service_password = TESTVALUES_REAL['rabbit_password'],
    messaging_service_urls = TESTVALUES_REAL['url_messaging_service'],
    messaging_service_exchange_name = TESTVALUES_REAL['messaging_exchange'],
    print_to_console = True,
    print_success_to_console = True
)
esgfpid.check_pid_queue_availability(**args)


if True:

    # Test variables:
    routing_key = TESTVALUES_REAL['routing_key']
    routing_key = 'BLABLA'
    num_messages = 5
    # Make connection:
    #testrabbit = self.make_rabbit_with_access()
    testrabbit = esgfpid.rabbit.rabbit.RabbitMessageSender(
          #exchange_name=TESTVALUES_REAL['exchange_name'],
          exchange_name='FOOEXCHANGE',
          urls_fallback=TESTVALUES_REAL['url_messaging_service'],
          url_preferred=None,
          username=TESTVALUES_REAL['rabbit_username'],
          password=TESTVALUES_REAL['rabbit_password'],
    )
    testrabbit.start()

    # Run code to be tested:
    for i in range(num_messages):
          print("Publishing message %i..." % i)
          testrabbit.send_message_to_queue({"stuffee":"foo"+str(i+1),"ROUTING_KEY":routing_key})
          time.sleep(1)
          print("Publishing message %i... done." % i)


    # Close connection after some time:
    time.sleep(60*5)
    testrabbit.finish()

    # Check result:
    leftovers = testrabbit.get_leftovers()
    if not len(leftovers)==0:
        print('There is leftovers: '+str(len(leftovers)))
        for message in leftovers:
            print('Leftover: '+str(message))
    if not testrabbit.is_finished():
        'Thread is still alive.'


import logging
import uuid
import json
import esgfpid.utils
import esgfpid.assistant.messages
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn


LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())

class ErrataAssistant(object):

    def __init__(self, **args):
        '''
        :param coupler: The coupler object (for sending the message to the queue).
        '''
        mandatory_args = ['prefix', 'coupler']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        esgfpid.utils.check_noneness_of_mandatory_args(args, mandatory_args)

        self.__prefix = args['prefix']
        self.__coupler = args['coupler']

    def add_errata_ids(self, **args):
        logdebug(LOGGER, 'Adding errata ids...')
        mandatory_args = ['drs_id', 'version_number', 'errata_ids']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        esgfpid.utils.check_noneness_of_mandatory_args(args, mandatory_args)

        dataset_handle = self.__get_dataset_handle(args)
        errata_ids = self.__get_errata_ids_as_list(args)
        message = self.__make_add_message(errata_ids, dataset_handle, args['drs_id'], args['version_number'])
        self.__send_message_to_queue(message)

        loginfo(LOGGER, 'Requesting to add errata ids "%s" to dataset "%s".', ', '.join(errata_ids), dataset_handle)
        logdebug(LOGGER, 'Adding errata ids... done')


    def remove_errata_ids(self, **args):
        logdebug(LOGGER, 'Removing errata ids...')
        mandatory_args = ['drs_id', 'version_number', 'errata_ids']
        esgfpid.utils.check_presence_of_mandatory_args(args, mandatory_args)
        esgfpid.utils.check_noneness_of_mandatory_args(args, mandatory_args)

        dataset_handle = self.__get_dataset_handle(args)
        errata_ids = self.__get_errata_ids_as_list(args)
        message = self.__make_remove_message(errata_ids, dataset_handle, args['drs_id'], args['version_number'])
        self.__send_message_to_queue(message)

        loginfo(LOGGER, 'Requesting to remove errata ids "%s" from dataset "%s".', ', '.join(errata_ids), dataset_handle)
        logdebug(LOGGER, 'Removing errata ids... done')

    def __get_errata_ids_as_list(self, args):
        errata_ids = args['errata_ids']
        if type(errata_ids) == type([]):
            return errata_ids
        else:
            return [errata_ids]

    def __get_dataset_handle(self, args):
        dataset_handle = esgfpid.utils.make_handle_from_drsid_and_versionnumber(
            drs_id=args['drs_id'],
            version_number=args['version_number'],
            prefix=self.__prefix
        )
        return dataset_handle

    def __make_add_message(self, errata_ids, dataset_handle, drs_id, vers_number):
        message_timestamp = esgfpid.utils.get_now_utc_as_formatted_string()
        
        message = esgfpid.assistant.messages.add_errata_ids_message(
            dataset_handle = dataset_handle,
            timestamp = message_timestamp,
            errata_ids = errata_ids,
            drs_id=drs_id,
            version_number=vers_number
        )
        return message

    def __make_remove_message(self, errata_ids, dataset_handle, drs_id, vers_number):
        message_timestamp = esgfpid.utils.get_now_utc_as_formatted_string()
        
        message = esgfpid.assistant.messages.remove_errata_ids_message(
            dataset_handle = dataset_handle,
            timestamp = message_timestamp,
            errata_ids = errata_ids,
            drs_id=drs_id,
            version_number=vers_number
        )
        return message


    def __send_message_to_queue(self, message):
        self.__coupler.send_message_to_queue(message)


import logging
import pika
import json
import esgfpid.assistant.messages
from esgfpid.utils import loginfo, logdebug, logtrace, logerror, logwarn, log_every_x_times

LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())


class UnacceptedMessagesHandler(object):

    def __init__(self, thread):
        self.thread = thread

        self.__have_warned_about_double_unroutable_already = False

    '''
    Callback, called by RabbitMQ.
    '''
    def on_message_not_accepted(self, channel, returned_frame, props, body):
        # Messages that are returned are confirmed anyways.
        # If we sent 20 messages that are returned, all 20 are acked,
        # so we do not need to retrieve them from the unconfirmed
        # messages after resending.
        # In the end, we'll have published 40 messages and received 40 acks.

        # Logging...
        logtrace(LOGGER, 'Return frame: %s', returned_frame) # <Basic.Return(['exchange=rabbitsender_integration_tests', 'reply_code=312', 'reply_text=NO_ROUTE', 'routing_key=cmip6.publisher.HASH.cart.datasets'])>
        logtrace(LOGGER, 'Return props: %s', props)  # <BasicProperties(['content_type=application/json', 'delivery_mode=2'])>
        logtrace(LOGGER, 'Return body: %s', body)

        # Was it the first or second time it comes back?
        if returned_frame.reply_text == 'NO_ROUTE':
            loginfo(LOGGER, 'The message was returned because it could not be assigned to any queue. No binding for routing key "%s".', returned_frame.routing_key)
            if returned_frame.routing_key.startswith(esgfpid.utils.RABBIT_EMERGENCY_ROUTING_KEY):
                self.__log_about_double_return(returned_frame, body)
            else:
                self.__resend_message(returned_frame, props, body)
        else:
            logerror(LOGGER, 'The message was returned. Routing key: %s. Unknown reason: %s', returned_frame.routing_key, returned_frame.reply_text)
            self.__resend_message(returned_frame, props, body)

    def __log_about_double_return(self, frame, body):
        if not self.__have_warned_about_double_unroutable_already:
            body_json = json.loads(body)
            logerror(LOGGER, 'The RabbitMQ node refused a message a second time (with the original routing key "%s" and the emergency routing key "%s"). Dropping the message.',
                body_json['original_routing_key'], frame.routing_key)
            self.__have_warned_about_double_unroutable_already = True
        logdebug(LOGGER, 'This is the second time the message comes back. Dropping it.')

    def __resend_message(self, returned_frame, props, body):
        try:
            body_json = json.loads(body)
            body_json = self.__add_emergency_routing_key(body_json)
            self.__resend_an_unroutable_message(json.dumps(body_json))
        except pika.exceptions.ChannelClosed as e:
            logdebug(LOGGER, 'Error during "on_message_not_accepted": %s: %s', e.__class__.__name__, repr(e))
            logerror(LOGGER, 'Could not resend message: %s: %s', e.__class__.__name__, repr(e))

    def __add_emergency_routing_key(self, body_json):
        emergency_routing_key = esgfpid.utils.RABBIT_EMERGENCY_ROUTING_KEY
        key_for_routing_key = esgfpid.assistant.messages.JSON_KEY_ROUTING_KEY

        # If there was no routing key, set the original one to 'None'
        if key_for_routing_key not in body_json:
            logerror(LOGGER, 'Very unexpected: RabbitMQ returned a message that had no routing key: %s', body_json)
            body_json[key_for_routing_key] = 'None'

        # If it already HAS the emergency routing key, do not adapt the routing key
        # (This means the message already came back a second time...)
        if body_json[key_for_routing_key] == emergency_routing_key:
            pass

        # Otherwise, store the original one in another field...
        # and overwrite it by the emergency routing key:
        else:
            body_json['original_routing_key'] = body_json[key_for_routing_key]
            logdebug(LOGGER, 'Adding emergency routing key %s', emergency_routing_key)
            body_json[key_for_routing_key] = emergency_routing_key
        return body_json

    def __resend_an_unroutable_message(self, message):
        logdebug(LOGGER, 'Resending message...')
        self.thread.send_a_message(message)


import mock
import copy
import esgfpid
import tests.resources.responsemock
#import tests.resources.solrmock
#import tests.resources.rabbitmock
import tests.utils as utils
import sys
if sys.version[0] == '2':
    import Queue as queue
else:
    import queue as queue

# Errata
ERRATA_SEVERAL = ['123456','654321']
ERRATA = '123456'

# Solr
SOLR_URL_LIBRARY = 'http://solr.foo'
SOLR_URL_CONSUMER = 'http://solr-to-be-used-by-consumer.foo'

# RabbitMQ
RABBIT_URL_TRUSTED = 'http://trusted.rabbit.foo'
RABBIT_URL_TRUSTED2 = 'http://trusted.rabbit.bar'
RABBIT_URL_TRUSTED3 = 'http://trusted.rabbit.baz'
RABBIT_URL_OPEN = 'http://open.rabbit.foo'
EXCHANGE_NAME = 'exch'
RABBIT_USER_OPEN = 'EvilRabbit'
RABBIT_USER_TRUSTED = 'HappyRabbit'
RABBIT_PASSWORD = 'carrotDreams'
TEST_RABBIT_CREDS_OPEN = dict(
    url = RABBIT_URL_OPEN,
    user = RABBIT_USER_OPEN)
TEST_RABBIT_CREDS_TRUSTED = dict(
    url = RABBIT_URL_TRUSTED,
    user = RABBIT_USER_TRUSTED,
    password = RABBIT_PASSWORD)

# Misc
PREFIX_NO_HDL = '21.14foo'
PREFIX_WITH_HDL = 'hdl:'+PREFIX_NO_HDL
PREFIX_FOR_ROUTINGKEY = esgfpid.utils.routingkeys._sanitize_prefix(PREFIX_NO_HDL)

# Files
SUFFIX_FILE = 'xyz123abc'
SUFFIX_FILE2 = 'xyz123def'
FILEHANDLE_HDL  = PREFIX_WITH_HDL+'/'+SUFFIX_FILE
FILEHANDLE_NO_HDL  = PREFIX_NO_HDL+'/'+SUFFIX_FILE
FILEHANDLE2_HDL = PREFIX_WITH_HDL+'/'+SUFFIX_FILE2
FILENAME = 'mytestfilename.nc'
CHECKSUM = 'testchecksumabc'
CHECKSUMTYPE = 'SHA999'
FILESIZE = 10010
FILEVERSION = '987'
PUBLISH_PATH = 'my/publish/path'

# Datasets
DRS_ID = 'my/drs/id'
THREDDS ='test/thr/path'
DATA_NODE = 'esgf-foo.dkrz.de'
SUFFIX_DS = 'afd65cd0-9296-35bc-a706-be98665c9c36'
SUFFIX_DS2 = 'ceb26955-82ea-3610-ab41-5ebfd1533f94'
DS_VERSION = 201603
DS_VERSION2 = 201703
DATASETHANDLE_HDL = PREFIX_WITH_HDL+'/'+SUFFIX_DS
DATASETHANDLE_HDL2 = PREFIX_WITH_HDL+'/'+SUFFIX_DS2


#
# Mocks
#


def get_thread_mock():
    return MockThread()

def get_thread_mock2(error=None):
    return MockThread2(error=error)

def get_thread_mock3(error=None):
    return MockThread3(error=error)

def get_thread_mock4(error=None):
    return MockThread4(error=error)

def get_connection_mock():
    connectionmock = mock.MagicMock()
    def side_effect_call_later(seconds, callback):
        callback()
    connectionmock.ioloop.call_later = mock.MagicMock()
    connectionmock.ioloop.call_later.side_effect = side_effect_call_later
    return connectionmock




class MockThread(object):

    def __init__(self):
        self.num_message_events = 0
        self.start_was_called = False
        self.was_gently_finished = False
        self.was_force_finished = False
        self._is_alive = False
        self.nacked = []
        self.unconfirmed = []
        self.messages = []

    # RabbitThread API:

    def start(self):
        self.start_was_called = True
        self._is_alive = True

    def join(self, seconds):
        self.join_was_called = True
        self._is_alive = False

    def add_event_publish_message(self):
        self.num_message_events += 1

    def add_event_force_finish(self):
        self.was_force_finished = True

    def add_event_gently_finish(self):
        self.was_gently_finished = True

    def is_alive(self):
        return self._is_alive

    def get_nacked_messages_as_list(self):
        return self.nacked

    def get_unconfirmed_messages_as_list_copy(self):
        return self.unconfirmed

'''
Used for testing the thread_feeder.
'''
class MockThread2(object):

    def __init__(self, error=None):
        # Used for mocking the behaviour:
        self.messages = []
        self.put_back = []
        self.undelivered_msg = []
        self.unconfirmed_tags = []
        self.exchange_name = 'foo'
        # Rabbit API, used by modules:
        self._channel = mock.MagicMock()
        if error is not None:
            self._channel.basic_publish.side_effect = error

    def get_message_from_unpublished_stack(self, seconds):
        if len(self.messages) == 0:
            raise queue.Empty()
        else:
            return self.messages.pop()

    def get_num_unpublished(self):
        return len(self.messages)

    def put_one_message_into_queue_of_unsent_messages(self, msg):
        self.messages.append(msg)
        self.put_back.append(msg)

    def get_exchange_name(self):
        return self.exchange_name

    def put_to_unconfirmed_delivery_tags(self, tag):
        self.unconfirmed_tags.append(tag)

    def put_to_unconfirmed_messages_dict(self, tag, msg):
        self.undelivered_msg.append(msg)


'''
Used for testing the thread_returner.
'''
class MockThread3(object):

    def __init__(self, error=None):
        # Attributes (used by modules):
        self._channel = mock.MagicMock()
        # Methods (used by modules):
        self.send_a_message = mock.MagicMock()

        if error is not None:
            self.send_a_message.side_effect = error

'''
Used for testing the thread_returner.
'''
class MockThread4(object):

    def __init__(self, error=None):
        self.num_unpublished = 0
        self.num_unconfirmed = 0.0
        self.fake_confirms = False
        self.confirm_rate = 1
        # Attributes (used by modules):
        self.ERROR_CODE_CONNECTION_FORCE_CLOSED = 999
        self.ERROR_TEXT_CONNECTION_FORCE_CLOSED = "errortext"
        self.ERROR_CODE_CONNECTION_CLOSED_BY_USER = 888
        self.ERROR_TEXT_CONNECTION_CLOSED_BY_USER = "bla"
        self.ERROR_CODE_CONNECTION_NORMAL_SHUTDOWN = 777
        self.ERROR_TEXT_CONNECTION_NORMAL_SHUTDOWN = 'bli'
        self._connection = mock.MagicMock()
        self._connection.is_closed = False
        self._connection.is_closing = False
        self._connection.is_open = True

        def side_effect_call_later(seconds, callback):
            callback()
        self._connection.ioloop.call_later = mock.MagicMock()
        self._connection.ioloop.call_later.side_effect = side_effect_call_later

        # Methods (used by modules):
        self.tell_publisher_to_stop_waiting_for_gentle_finish = mock.MagicMock()
        self.make_permanently_closed_by_user = mock.MagicMock()
        self.add_event_publish_message = mock.MagicMock()
        self.add_event_publish_message.side_effect = self.side_effect_add_event_publish_message

    def get_num_unpublished(self):
        #print('Unpublished: '+str(self.num_unpublished))
        return int(round(self.num_unpublished))

    def get_num_unconfirmed(self):
        num_conf = self.num_unconfirmed
        # This fakes receiving confirms from Rabbit:
        if self.fake_confirms:
            self.num_unconfirmed = self.num_unconfirmed - self.confirm_rate
            if (self.num_unconfirmed < 0):
                self.num_unconfirmed = 0
        # This returns num confirms
        #print('Unconfirmed: '+str(self.num_unconfirmed))
        return num_conf

    def side_effect_add_event_publish_message(self):
        self.num_unpublished = self.num_unpublished -1
        if (self.num_unpublished < 0):
            self.num_unpublished = 0
        #print('Called add_event_publish_message! Now: %i' % self.num_unpublished)

    def set_fake_confirms(self):
        self.fake_confirms = True






class MockRabbitSender(object):

    def __init__(self):
        self.received_messages = []
        self.please_print = False

    def send_message_to_queue(self, msg):
        self.received_messages.append(msg)
        if self.please_print:
            print('Called "send_message_to_queue()" with '+str(msg))

    def open_rabbit_connection(self):
        if self.please_print:
            print('Called "open_rabbit_connection()"')

    def close_rabbit_connection(self):
        if self.please_print:
            print('Called "close_rabbit_connection()"')

    def start(self):
        if self.please_print:
            print('Called "start()"')

    def finish(self):
        if self.please_print:
            print('Called "finish()"')

    def force_finish(self):
        if self.please_print:
            print('Called "force_finish()"')

#
# Provide a coupler for unit tests
# that can be mocked with solr mock, rabbit mock...
#

def get_coupler_args(**kwargs):
    cred_copy = copy.deepcopy(TEST_RABBIT_CREDS_TRUSTED)
    coupler_args = dict(
        handle_prefix = PREFIX_NO_HDL,
        messaging_service_credentials = [cred_copy],
        messaging_service_exchange_name = EXCHANGE_NAME,
        data_node = DATA_NODE,
        thredds_service_path = THREDDS,
        solr_url = SOLR_URL_LIBRARY,
        solr_https_verify=True,
        solr_switched_off=False,
        test_publication=False,
        message_service_synchronous=False
    )
    for k,v in kwargs.items():
        coupler_args[k] = v
    return coupler_args

def get_coupler(**kwargs):
    coupler_args = get_coupler_args(**kwargs)
    testcoupler = esgfpid.coupling.Coupler(**coupler_args)
    return testcoupler


#
# Provide a connector
# that can be mocked with solr mock, rabbit mock...
#

def get_rabbit_credentials(**kwargs):
    rabbit_args = dict(
        user = RABBIT_USER_TRUSTED,
        url = RABBIT_URL_TRUSTED,
        password = RABBIT_PASSWORD
    )
    for k,v in kwargs.items():
        rabbit_args[k] = v
    return rabbit_args

def get_connector_with_rabbit(**kwargs):
    cred = get_rabbit_credentials(**kwargs)
    return get_connector(messaging_service_credentials=[cred])

def get_connector_args(**kwargs):
    cred_copy = copy.deepcopy(TEST_RABBIT_CREDS_TRUSTED)
    connector_args = dict(
        handle_prefix = PREFIX_NO_HDL,
        messaging_service_exchange_name = EXCHANGE_NAME,
        messaging_service_credentials = [cred_copy]
    )
    for k,v in kwargs.items():
        connector_args[k] = v
    return connector_args

def get_connector(**kwargs):
    connector_args = get_connector_args(**kwargs)
    return esgfpid.Connector(**connector_args)

#
# Patching RabbitMQ
#

def patch_with_rabbit_mock(connector_or_coupler, rabbitmock=None):
    if rabbitmock is None:
        rabbitmock = MockRabbitSender()
    if type(connector_or_coupler) == esgfpid.Connector:
        connector_or_coupler._Connector__coupler._Coupler__rabbit_message_sender = rabbitmock
    else:
        connector_or_coupler._Coupler__rabbit_message_sender = rabbitmock
    return rabbitmock

def get_received_message_from_rabbitmock(connector_or_coupler, index=0):
    if type(connector_or_coupler) == esgfpid.Connector:
        return _get_received_message_from_connector_rabbitmock(connector_or_coupler, index)
    else:
        return _get_received_message_from_coupler_rabbitmock(connector_or_coupler, index)

def _get_received_message_from_connector_rabbitmock(connector, index=0):
    coupler = connector._Connector__coupler
    rabbitmock = coupler._Coupler__rabbit_message_sender
    msg = rabbitmock.received_messages[index] #  if using MockRabbitSender
    #msg = rabbitmock.send_message_to_queue.call_args[0][index] # if using MagicMock (first get positional args, then get the i'th of those)
    tests.utils.replace_date_with_string(msg)
    return msg

def _get_received_message_from_coupler_rabbitmock(coupler, index=0):
    rabbitmock = coupler._Coupler__rabbit_message_sender
    msg = rabbitmock.received_messages[index] # if using MockRabbitSender
    #msg = rabbitmock.send_message_to_queue.call_args[0][index] # if using MagicMock (first get positional args, then get the i'th of those)
    tests.utils.replace_date_with_string(msg)
    return msg

#
# Patching solr
#

def patch_solr_returns_empty_file_list(connector_or_coupler):
    return patch_solr_returns_previous_files(connector_or_coupler, [])

def patch_solr_returns_previous_files(connector_or_coupler, previous_files):
    solrmock = mock.Mock()
    solrmock.retrieve_file_handles_of_same_dataset = mock.Mock()
    solrmock.retrieve_file_handles_of_same_dataset.return_value = previous_files
    patch_with_solr_mock(connector_or_coupler, solrmock)

def patch_solr_returns_list_of_datasets_and_versions(connector_or_coupler, datasets, versions):
    solrmock = mock.Mock()
    methodmock = mock.Mock()
    methodmock.return_value = dict(dataset_handles=datasets, version_numbers=versions)
    solrmock.retrieve_datasethandles_or_versionnumbers_of_allversions = methodmock
    patch_with_solr_mock(connector_or_coupler, solrmock)

def patch_solr_raises_error(connector_or_coupler, error_to_be_raised):
    solrmock = mock.Mock()
    methodmock = mock.Mock()
    methodmock.side_effect = error_to_be_raised
    solrmock.retrieve_file_handles_of_same_dataset = methodmock
    solrmock.retrieve_datasethandles_or_versionnumbers_of_allversions = methodmock
    patch_with_solr_mock(connector_or_coupler, solrmock)

def patch_with_solr_mock(connector_or_coupler, solrmock):
    solrmock.is_switched_off = mock.Mock()
    solrmock.is_switched_off.return_value = False
    if type(connector_or_coupler) == esgfpid.Connector:
        connector_or_coupler._Connector__coupler._Coupler__solr_sender = solrmock
    else:
        connector_or_coupler._Coupler__solr_sender = solrmock




#####

def get_rabbit_args(**kwargs):
    if not 'is_synchronous_mode' in kwargs:
        raise ValueError('Please specify argument "is_synchronous_mode"!')
    cred_copy = copy.deepcopy(TEST_RABBIT_CREDS_TRUSTED)
    rabbit_args = dict(
        credentials = [cred_copy],
        exchange_name='exch',
        test_publication=False,
        is_synchronous_mode=kwargs['is_synchronous_mode']
    )
    for k,v in kwargs.items():
        rabbit_args[k] = v
    return rabbit_args

def get_rabbit_message_sender(**kwargs):
    args = get_rabbit_args(**kwargs)
    testrabbit = esgfpid.rabbit.RabbitMessageSender(**args)
    return testrabbit

def patch_rabbit_with_magic_mock(testrabbit):
    testrabbit._RabbitMessageSender__server_connector = mock.MagicMock()

def get_nodemanager():
    nodemanager = esgfpid.rabbit.nodemanager.NodeManager()
    nodemanager.add_trusted_node(
        username=RABBIT_USER_TRUSTED,
        host = RABBIT_URL_TRUSTED,
        exchange_name = EXCHANGE_NAME,
        password=RABBIT_PASSWORD,
        priority=1)
    nodemanager.add_trusted_node(
        username=RABBIT_USER_TRUSTED,
        host = RABBIT_URL_TRUSTED2,
        exchange_name = EXCHANGE_NAME,
        password=RABBIT_PASSWORD,
        priority=2)
    nodemanager.add_trusted_node(
        username=RABBIT_USER_TRUSTED,
        host = RABBIT_URL_TRUSTED3,
        exchange_name = EXCHANGE_NAME,
        password=RABBIT_PASSWORD,
        priority=3)
    return nodemanager

def get_synchronous_rabbit():
    return esgfpid.rabbit.synchronous.SynchronousRabbitConnector(get_nodemanager())

def get_asynchronous_rabbit():
    return esgfpid.rabbit.asynchronous.AsynchronousRabbitConnector(get_nodemanager())


#
# Solr
#

def get_testsolr(**kwargs):
    solr_args = dict(
        solr_url = SOLR_URL_LIBRARY,
        prefix = PREFIX_NO_HDL,
        switched_off = False,
        https_verify = True,
        disable_insecure_request_warning = False
    )
    for k,v in kwargs.items():
        solr_args[k] = v
    testsolr = esgfpid.solr.SolrInteractor(**solr_args)
    return testsolr

def get_testsolr_switched_off():
    return get_testsolr(switched_off=True)

def get_testsolr_connector():
    testsolr = esgfpid.solr.serverconnector.SolrServerConnector(
        solr_url = SOLR_URL_LIBRARY,
        https_verify = True,
        disable_insecure_request_warning = False
    )
    return testsolr

#
# Args for common tasks...
#

def get_args_for_publication_assistant():
    # Coupler has to be added!
    return dict(
        prefix = PREFIX_NO_HDL,
        drs_id = DRS_ID,
        version_number=DS_VERSION,
        is_replica=False,
        thredds_service_path = THREDDS,
        data_node = DATA_NODE,
        consumer_solr_url = SOLR_URL_CONSUMER
    )

def get_args_for_adding_file():
    return dict(
        file_name = FILENAME,
        file_handle = PREFIX_WITH_HDL+'/'+SUFFIX_FILE,
        checksum = CHECKSUM,
        file_size = FILESIZE,
        publish_path = PUBLISH_PATH,
        prefix = PREFIX_NO_HDL,
        checksum_type = CHECKSUMTYPE,
        file_version = FILEVERSION
    )

def get_args_for_unpublish_one():
    # Coupler has to be added!
    return dict(
        drs_id=DRS_ID,
        prefix=PREFIX_NO_HDL,
        data_node=DATA_NODE)

def get_args_for_unpublish_all():
    # Coupler has to be added!
    return dict(
        drs_id=DRS_ID,
        prefix=PREFIX_NO_HDL,
        data_node=DATA_NODE)

def get_args_for_consistency_check():
    # Coupler has to be added!
    return dict(
        drs_id = DRS_ID,
        data_node = DATA_NODE,
        version_number = DS_VERSION)

def get_args_for_nodemanager(**kwargs):
    args = dict(
        exchange_name='exch_foo',
        host='host_foo',
        username='user_foo',
        password='pw_foo'
    )
    for k,v in kwargs.items():
        args[k] = v
    return args

#
# Rabbit Messages
#

def get_rabbit_message_publication_dataset():
    expected_rabbit_task = {
        "handle" : DATASETHANDLE_HDL,
        "aggregation_level" : "dataset",
        "operation" : "publish",
        "message_timestamp" : "anydate",
        "drs_id" : DRS_ID,
        "version_number" : DS_VERSION,
        "files" : [FILEHANDLE_HDL],
        "is_replica" : False,
        "data_node" : DATA_NODE,
        "ROUTING_KEY" : PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.publi-ds-orig',
        "consumer_solr_url" : SOLR_URL_CONSUMER
    }
    return expected_rabbit_task

def get_rabbit_message_publication_file():
    expected_data_url = ('http://'+DATA_NODE+'/'+THREDDS+'/'+PUBLISH_PATH)
    expected_rabbit_task = {
        "handle": FILEHANDLE_HDL,
        "aggregation_level": "file",
        "operation": "publish",
        "message_timestamp": "anydate",
        "is_replica": False,
        "file_name": FILENAME,
        "data_url": expected_data_url,
        "data_node": DATA_NODE,
        "file_size": FILESIZE,
        "checksum": CHECKSUM,
        "checksum_type": CHECKSUMTYPE,
        "file_version": FILEVERSION,
        "parent_dataset": DATASETHANDLE_HDL,
        "ROUTING_KEY" : PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.publi-file-orig',
    }
    return expected_rabbit_task

def get_rabbit_message_unpub_one():
    expected_rabbit_task = {
        "handle": DATASETHANDLE_HDL,
        "aggregation_level": "dataset",
        "operation": "unpublish_one_version",
        "message_timestamp":"anydate",
        "data_node": DATA_NODE,
        "ROUTING_KEY" : PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.unpubli-onevers',
        "drs_id":DRS_ID,
        "version_number":int(DS_VERSION)
    }
    return expected_rabbit_task

def get_rabbit_message_unpub_one_by_handle():
    expected_rabbit_task = {
        "handle": DATASETHANDLE_HDL,
        "aggregation_level": "dataset",
        "operation": "unpublish_one_version",
        "message_timestamp":"anydate",
        "data_node": DATA_NODE,
        "ROUTING_KEY" : PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.unpubli-onevers',
        "drs_id": None,
    }
    return expected_rabbit_task

def get_rabbit_message_unpub_all():
    expected_rabbit_task = {
        "operation": "unpublish_all_versions",
        "aggregation_level": "dataset",
        "message_timestamp": "anydate",
        "drs_id": DRS_ID,
        "data_node": DATA_NODE,
        "ROUTING_KEY" : PREFIX_FOR_ROUTINGKEY+'.HASH.fresh.unpubli-allvers',
    }
    return expected_rabbit_task
